import os
from pathlib import Path

from PIL import Image, ImageSequence

import pillow_heif.HeifImagePlugin  # noqa

os.chdir(os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))), "tests"))
TARGET_FOLDER = "../converted"

# zPug_3 contains three images, we remove first image
if __name__ == "__main__":
    os.makedirs(TARGET_FOLDER, exist_ok=True)
    image_path = Path("images/heif/zPug_3.heic")
    pillow_img = Image.open(image_path)
    result_path = os.path.join(TARGET_FOLDER, f"{image_path.stem}.heic")
    frames = []
    for frame in ImageSequence.Iterator(pillow_img):
        frames.append(frame.copy())
    del frames[0]
    frames[0].save(result_path, save_all=True, quality=35, append_images=frames[1:])



"""
Functions and classes for heif images to read and write.
"""

from typing import Any, Dict, Iterator, List, Union
from warnings import warn
from weakref import ref

from _pillow_heif_cffi import ffi, lib
from PIL import Image, ImageOps, ImageSequence

from ._lib_info import have_encoder_for_format
from ._libheif_ctx import LibHeifCtx, LibHeifCtxWrite
from ._options import options
from .constants import (
    HeifChannel,
    HeifChroma,
    HeifColorspace,
    HeifCompressionFormat,
    HeifFiletype,
)
from .error import HeifError, HeifErrorCode, check_libheif_error
from .misc import _get_bytes, set_orientation
from .private import (
    FFI_DRY_ALLOC,
    MODE_CONVERT,
    MODE_INFO,
    HeifCtxAsDict,
    exif_from_pillow,
    get_pure_stride,
    read_color_profile,
    read_metadata,
    retrieve_exif,
    retrieve_xmp,
    set_color_profile,
    set_exif,
    set_metadata,
    set_xmp,
    xmp_from_pillow,
)


class HeifImageBase:
    """Base class for :py:class:`HeifImage` and :py:class:`HeifThumbnail`"""

    size: tuple
    """Width and height of the image."""

    mode: str
    """A string which defines the type and depth of a pixel in the image:
    `Pillow Modes <https://pillow.readthedocs.io/en/stable/handbook/concepts.html#modes>`_
    For currently supported modes by Pillow-Heif see :ref:`image-modes`."""

    def __init__(self, heif_ctx: Union[LibHeifCtx, HeifCtxAsDict], handle, for_encoding=False):
        self._img_data: Dict[str, Any] = {}
        self._heif_ctx = heif_ctx
        if isinstance(heif_ctx, LibHeifCtx):
            self._handle = ffi.gc(handle, lib.heif_image_handle_release)
            self.size = (
                lib.heif_image_handle_get_width(self._handle),
                lib.heif_image_handle_get_height(self._handle),
            )
            self.mode = "RGBA" if bool(lib.heif_image_handle_has_alpha_channel(self._handle)) else "RGB"
            bit_depth = 8 if heif_ctx.to_8bit else lib.heif_image_handle_get_luma_bits_per_pixel(self._handle)
            if bit_depth > 8:
                self.mode += f";{bit_depth}"
        else:
            self._handle = None
            self.size = heif_ctx.size
            self.mode = heif_ctx.mode
            if heif_ctx.data:
                new_mode = MODE_INFO[self.mode][4] if for_encoding else None
                if isinstance(new_mode, tuple):
                    new_mode = new_mode[1] if options().save_to_12bit else new_mode[0]
                _img = self._create_image(heif_ctx.data, heif_ctx.stride, new_mode)
                self._img_to_img_data_dict(_img)

    @property
    def bit_depth(self):
        """Image channel pixel bit depth. Possible values: 8, 10, 12, 16

        .. note:: When ``convert_hdr_to_8bit`` is True, return value will be always ``8`` when
         image opened from file(e.g. was not created with ``from_pillow`` or ``add_from_*`` functions)."""

        return MODE_INFO[self.mode][1]

    @property
    def original_bit_depth(self):
        """Number of bits in colour channel, before it was decoded using ``convert_hdr_to_8bit`` parameter.

        .. note:: If ``convert_hdr_to_8bit`` is ``False`` then this field is always equal to ``bit_depth``

            This includes situations where image was created not from a file.

        :returns: An int value representing number of bits per color channel."""

        if self._handle is not None:
            return lib.heif_image_handle_get_luma_bits_per_pixel(self._handle)
        return self.bit_depth

    @property
    def has_alpha(self):
        """``True`` for images with ``alpha`` channel, ``False`` otherwise.

        :returns: "True" or "False" """

        return self.mode.find("A") != -1

    @property
    def heif_img(self):
        self._load_if_not()
        return self._img_data.get("img", None)

    @property
    def data(self):
        """Decodes image and returns image data from ``libheif``. See :ref:`image_data`

        .. note:: Actual size of data returned by ``data`` can be bigger then ``width * height * pixel size``.

        :returns: ``bytes`` of the decoded image from ``libheif``."""

        self._load_if_not()
        return self._img_data.get("data", None)

    @property
    def stride(self):
        """Decodes image and returns stride.

        :returns: ``int`` stride of the decoded image."""

        self._load_if_not()
        return self._img_data.get("stride", None)

    def convert_to(self, mode: str) -> None:
        """Decode and convert in place image to the specified mode.

        :param mode: for list of supported conversions, see: :ref:`convert_to`

        :exception KeyError: If conversion between modes is not supported."""

        current_stride = self.stride
        current_data = bytes(self.data)  # copying `data` to bytes so it will not be GC collected.
        self.unload()
        _img = self._create_image(current_data, current_stride, mode)
        self._img_to_img_data_dict(_img)

    def to_pillow(self) -> Image.Image:
        """Helper method to create :py:class:`PIL.Image.Image`

        :returns: :py:class:`PIL.Image.Image` class created from this image/thumbnail."""

        image = Image.frombytes(
            self.mode,  # noqa
            self.size,
            self.data,
            "raw",
            self.mode,
            self.stride,
        )
        info = None
        if isinstance(self, HeifImage):
            image.info["thumbnails"] = [i.clone_nd() for i in getattr(self, "thumbnails", [])]
            info = self.info
        elif isinstance(self, HeifThumbnail):
            _ = self.get_original()
            info = _.info if _ else None
        if info:
            for k in ("exif", "xmp", "metadata", "primary", "icc_profile", "icc_profile_type", "nclx_profile"):
                if k in info:
                    image.info[k] = info[k]
            image.info["original_orientation"] = set_orientation(image.info)
        return image

    def load(self):
        """Decode image.

        Usually, you do not need to call this, image will be decoded automatically
        when accessing ``data`` or ``stride`` properties."""

        self._load_if_not()
        return self

    def unload(self):
        if self._handle is not None:
            self._img_data.clear()

    @property
    def __array_interface__(self):
        """Numpy array interface support"""

        if isinstance(self, HeifThumbnail) and self.data is None:
            raise ValueError("Usage error: thumbnail has no data.")
        shape = (self.size[1], self.size[0])
        if MODE_INFO[self.mode][0] > 1:
            shape += (MODE_INFO[self.mode][0],)
        typestr = MODE_INFO[self.mode][5]
        data = bytes(self._get_pure_data())
        return {"shape": shape, "typestr": typestr, "version": 3, "data": data}

    def _color(self) -> HeifColorspace:
        return MODE_INFO[self.mode][2]

    def _chroma(self) -> HeifChroma:
        return MODE_INFO[self.mode][3]

    def _channel(self) -> HeifChannel:
        return HeifChannel.Y if self._color() == HeifColorspace.MONOCHROME else HeifChannel.INTERLEAVED

    def _load_if_not(self):
        if self._img_data or self._handle is None:
            return
        p_options = lib.heif_decoding_options_alloc()
        p_options = ffi.gc(p_options, lib.heif_decoding_options_free)
        p_options.convert_hdr_to_8bit = int(self._heif_ctx.to_8bit)
        p_img = ffi.new("struct heif_image **")
        check_libheif_error(lib.heif_decode_image(self._handle, p_img, self._color(), self._chroma(), p_options))
        heif_img = ffi.gc(p_img[0], lib.heif_image_release)
        self._img_to_img_data_dict(heif_img)

    def _img_to_img_data_dict(self, heif_img):
        p_stride = ffi.new("int *")
        p_data = lib.heif_image_get_plane(heif_img, self._channel(), p_stride)
        stride = p_stride[0]
        data_length = self.size[1] * stride
        data_buffer = ffi.buffer(p_data, data_length)
        self._img_data.update(img=heif_img, data=data_buffer, stride=stride)

    def _create_image(self, src_data, src_stride: int, new_mode=None):
        width, height = self.size
        p_new_img = ffi.new("struct heif_image **")
        chroma = MODE_INFO[new_mode][3] if new_mode else self._chroma()
        error = lib.heif_image_create(width, height, self._color(), chroma, p_new_img)
        check_libheif_error(error)
        new_img = ffi.gc(p_new_img[0], lib.heif_image_release)
        depth = MODE_INFO[new_mode][1] if new_mode else self.bit_depth
        error = lib.heif_image_add_plane(new_img, self._channel(), width, height, depth)
        check_libheif_error(error)
        p_dest_stride = ffi.new("int *")
        dest_data = lib.heif_image_get_plane(new_img, self._channel(), p_dest_stride)
        dest_stride = p_dest_stride[0]
        src_data = ffi.from_buffer(src_data)
        if new_mode and new_mode != self.mode:
            MODE_CONVERT[self.mode][new_mode](src_data, src_stride, dest_data, dest_stride, height)
            self.mode = new_mode
        else:
            lib.copy_image_data(src_data, src_stride, dest_data, dest_stride, height)
        return new_img

    def _get_pure_data(self):
        new_stride = get_pure_stride(self.mode, self.size[0])
        new_size = new_stride * self.size[1]
        new_data = FFI_DRY_ALLOC("char[]", new_size)
        lib.copy_image_data(ffi.from_buffer(self.data), self.stride, new_data, new_stride, self.size[1])
        return ffi.buffer(new_data, new_size)


class HeifThumbnail(HeifImageBase):
    """Class represents a single thumbnail for a HeifImage."""

    def __init__(self, original_img, reference, thumb_id: int = None):
        if isinstance(original_img, HeifImage):
            p_handle = ffi.new("struct heif_image_handle **")
            check_libheif_error(lib.heif_image_handle_get_thumbnail(original_img._handle, thumb_id, p_handle))
            super().__init__(original_img._heif_ctx, p_handle[0])
        else:
            super().__init__(original_img, None)
        self.parent = ref(reference)

    def __repr__(self):
        _bytes = f"{len(self.data)} bytes" if self._img_data else "no"
        return (
            f"<{self.__class__.__name__} {self.size[0]}x{self.size[1]} {self.mode} "
            f"with {_bytes} image data> Original:{self.parent()}"
        )

    def get_original(self):
        """Return an :py:class:`~pillow_heif.HeifImage` frame to which thumbnail belongs.

        .. note:: It is useful when you traverse thumbnails with :py:meth:`~pillow_heif.HeifFile.thumbnails_all`

        :returns: :py:class:`~pillow_heif.HeifImage` or None."""

        referenced = self.parent()
        return referenced if isinstance(referenced, HeifImage) else None

    def clone_nd(self, ref_original=None):
        """Private. Cloning thumbnail without actual data."""

        heif_ctx = HeifCtxAsDict(self.mode, self.size, None, stride=0)
        return HeifThumbnail(heif_ctx, ref_original if ref_original else heif_ctx)


class HeifImage(HeifImageBase):
    """Class represents one frame in a :py:class:`~pillow_heif.HeifFile`"""

    def __init__(self, heif_ctx: Union[LibHeifCtx, HeifCtxAsDict], img_id=-1, primary=False, for_encoding=False):
        additional_info = {}
        if isinstance(heif_ctx, LibHeifCtx):
            p_handle = ffi.new("struct heif_image_handle **")
            if primary:
                error = lib.heif_context_get_primary_image_handle(heif_ctx.ctx, p_handle)
            else:
                error = lib.heif_context_get_image_handle(heif_ctx.ctx, img_id, p_handle)
            check_libheif_error(error)
            handle = p_handle[0]
            _metadata = read_metadata(handle)
            _exif = retrieve_exif(_metadata)
            _xmp = retrieve_xmp(_metadata)
            additional_info["metadata"] = _metadata
            _color_profile = read_color_profile(handle)
            if _color_profile:
                if _color_profile["type"] in ("rICC", "prof"):
                    additional_info["icc_profile"] = _color_profile["data"]
                    additional_info["icc_profile_type"] = _color_profile["type"]
                else:
                    additional_info["nclx_profile"] = _color_profile["data"]
        else:
            handle = None
            _exif = None
            _xmp = None
            additional_info["metadata"] = []
            additional_info.update(heif_ctx.additional_info)
        super().__init__(heif_ctx, handle, for_encoding)
        self.info = {
            "exif": _exif,
            "xmp": _xmp,
            "primary": primary,
        }
        self.info.update(**additional_info)
        self.thumbnails = self.__read_thumbnails()

    def __repr__(self):
        _bytes = f"{len(self.data)} bytes" if self._img_data else "no"
        return (
            f"<{self.__class__.__name__} {self.size[0]}x{self.size[1]} {self.mode} "
            f"with {_bytes} image data and {len(self.thumbnails)} thumbnails>"
        )

    def scale(self, width: int, height: int):
        """Rescale image by a specific width and height given in parameters.

        .. note:: Image will be scaled in place. Images converted to some specific modes not always can be scaled.

        :param width: new image width.
        :param height: new image height."""

        self._load_if_not()
        p_scaled_img = ffi.new("struct heif_image **")
        check_libheif_error(lib.heif_image_scale_image(self.heif_img, p_scaled_img, width, height, ffi.NULL))
        scaled_heif_img = ffi.gc(p_scaled_img[0], lib.heif_image_release)
        self.size = (
            lib.heif_image_get_primary_width(scaled_heif_img),
            lib.heif_image_get_primary_height(scaled_heif_img),
        )
        self._img_to_img_data_dict(scaled_heif_img)
        return self

    def copy_thumbnails(self, thumbnails: List[HeifThumbnail]):
        """Private. For use only in ``add_from_pillow`` and ``add_from_heif``."""

        for thumb in thumbnails:
            self.thumbnails.append(thumb.clone_nd(ref_original=self))

    def __read_thumbnails(self) -> List[HeifThumbnail]:
        result: List[HeifThumbnail] = []
        if self._handle is None or not options().thumbnails:
            return result
        thumbs_count = lib.heif_image_handle_get_number_of_thumbnails(self._handle)
        if thumbs_count == 0:
            return result
        thumbnails_ids = ffi.new("heif_item_id[]", thumbs_count)
        thumb_count = lib.heif_image_handle_get_list_of_thumbnail_IDs(self._handle, thumbnails_ids, thumbs_count)
        for i in range(thumb_count):
            result.append(HeifThumbnail(self, self, thumb_id=thumbnails_ids[i]))
        return result


class HeifFile:
    """This class represents the :py:class:`~pillow_heif.HeifImage` classes container.

    To create :py:class:`~pillow_heif.HeifFile` object, use the appropriate factory functions.

    * :py:func:`~pillow_heif.open_heif`
    * :py:func:`~pillow_heif.from_pillow`
    * :py:func:`~pillow_heif.from_bytes`

    .. note:: To get empty container to fill it later, create a class without parameters."""

    def __init__(
        self, heif_ctx: Union[LibHeifCtx, HeifCtxAsDict] = None, img_ids: List[int] = None, main_id: int = None
    ):
        if heif_ctx is None:
            heif_ctx = HeifCtxAsDict("", (0, 0), None)
        self.images: List[HeifImage] = []
        self.mimetype = heif_ctx.get_mimetype() if isinstance(heif_ctx, LibHeifCtx) else ""
        if img_ids:
            for img_id in img_ids:
                self.images.append(HeifImage(heif_ctx, img_id, img_id == main_id))

    @property
    def original_bit_depth(self):
        """Points to :py:attr:`~pillow_heif.HeifImage.original_bit_depth` property of the
        primary :py:class:`~pillow_heif.HeifImage` in the container.

        :exception IndexError: If there is no images."""

        return self.images[self.primary_index()].original_bit_depth

    @property
    def bit_depth(self):
        """Points to :py:attr:`~pillow_heif.HeifImage.bit_depth` property of the
        primary :py:class:`~pillow_heif.HeifImage` in the container.

        :exception IndexError: If there is no images."""

        return self.images[self.primary_index()].bit_depth

    @property
    def size(self):
        """Points to :py:attr:`~pillow_heif.HeifImage.size` property of the
        primary :py:class:`~pillow_heif.HeifImage` in the container.

        :exception IndexError: If there is no images."""

        return self.images[self.primary_index()].size

    @property
    def mode(self):
        """Points to :py:attr:`~pillow_heif.HeifImage.mode` property of the
        primary :py:class:`~pillow_heif.HeifImage` in the container.

        :exception IndexError: If there is no images."""

        return self.images[self.primary_index()].mode

    @property
    def data(self):
        """Points to :py:attr:`~pillow_heif.HeifImage.data` property of the
        primary :py:class:`~pillow_heif.HeifImage` in the container.

        :exception IndexError: If there is no images."""

        return self.images[self.primary_index()].data

    @property
    def stride(self):
        """Points to :py:attr:`~pillow_heif.HeifImage.stride` property of the
        primary :py:class:`~pillow_heif.HeifImage` in the container.

        :exception IndexError: If there is no images."""

        return self.images[self.primary_index()].stride

    @property
    def has_alpha(self):
        """Points to :py:attr:`~pillow_heif.HeifImage.has_alpha` property of the
        primary :py:class:`~pillow_heif.HeifImage` in the container.

        :exception IndexError: If there is no images."""

        return self.images[self.primary_index()].has_alpha

    @property
    def info(self):
        """Points to ``info`` dict of the primary :py:class:`~pillow_heif.HeifImage` in the container.

        :exception IndexError: If there is no images."""

        return self.images[self.primary_index()].info

    @property
    def thumbnails(self):
        """Points to ``thumbnails`` of the primary :py:class:`~pillow_heif.HeifImage` in the container.

        :exception IndexError: If there is no images."""

        return self.images[self.primary_index()].thumbnails

    def primary_index(self, image_list=None) -> int:
        """Returns index of the ``PrimaryImage`` in the container."""

        if image_list is None:
            image_list = self.images
        i = 0
        for index, _ in enumerate(image_list):
            if _.info["primary"]:
                i = index
        return i

    def thumbnails_all(self, one_for_image: bool = False) -> Iterator[HeifThumbnail]:
        """Enums all thumbnails in all images.

        :param one_for_image: If set to ``True`` will return maximum one thumbnail for one image.

        :returns: Iterator for :py:class:`~pillow_heif.HeifThumbnail` classes."""

        for i in self:
            for thumb in i.thumbnails:
                yield thumb
                if one_for_image:
                    break

    def load(self, everything: bool = False):
        if everything:
            for img in self:
                img.load()
        else:
            self.images[self.primary_index()].load()
        return self

    def convert_to(self, mode: str) -> None:
        """Decode and convert in place primary image to the specified mode.

        :param mode: for list of supported conversions, see: :ref:`convert_to`

        :exception KeyError: If conversion between modes is not supported."""

        self.images[self.primary_index()].convert_to(mode)

    def scale(self, width: int, height: int) -> None:
        """Scale primary image in the container. See :py:meth:`~pillow_heif.HeifImage.scale`"""

        self.images[self.primary_index()].scale(width, height)

    def add_from_pillow(self, pil_image: Image.Image, load_one=False, ignore_primary=True, **kwargs):
        """Add image(s) to container.

        :param pil_image: ``PIL.Image`` class to get images from.
        :param load_one: should be only one frame loaded. Default=``False``
        :param ignore_primary: force ``PrimaryImage=False`` flag to all added images."""

        if load_one:
            self.__add_frame_from_pillow(pil_image, ignore_primary, **kwargs)
        else:
            current_frame = pil_image.tell() if hasattr(pil_image, "tell") else None
            for frame in ImageSequence.Iterator(pil_image):
                self.__add_frame_from_pillow(frame, ignore_primary, **kwargs)
            if current_frame is not None and hasattr(pil_image, "seek"):
                pil_image.seek(current_frame)
        return self

    def __add_frame_from_pillow(self, frame: Image.Image, ignore_primary: bool, **kwargs) -> None:
        if frame.width <= 0 or frame.height <= 0:
            return
        additional_info = {}
        supported_info_keys = (
            "exif",
            "xmp",
            "metadata",
            "primary",
            "icc_profile",
            "icc_profile_type",
            "nclx_profile",
        )
        for k in supported_info_keys:
            if k in frame.info:
                additional_info[k] = frame.info[k]
        if ignore_primary:
            additional_info["primary"] = False
        exif_from_pillow(additional_info, frame)
        xmp_from_pillow(additional_info, frame)
        original_orientation = set_orientation(additional_info)
        if frame.mode == "P":
            mode = "RGBA" if frame.info.get("transparency") else "RGB"
            frame = frame.convert(mode=mode)
        elif frame.mode == "LA":  # libheif doesnt not support INTERLEAVED_MONOCHROME mode
            frame = frame.convert(mode="RGBA")
        elif frame.mode == "I":
            frame = frame.convert(mode="I;16L")
        elif frame.mode == "1":
            frame = frame.convert(mode="L")

        if original_orientation is not None and original_orientation != 1:
            frame = ImageOps.exif_transpose(frame)
        added_image = self.add_frombytes(
            frame.mode,
            frame.size,
            frame.tobytes(),
            add_info={**additional_info},
            for_encoding=kwargs.get("for_encoding", False),
        )
        added_image.copy_thumbnails(frame.info.get("thumbnails", []))

    def add_from_heif(self, heif_image, load_one=False, ignore_primary=True, **kwargs):
        """Add image(s) to container.

        :param heif_image: ``HeifFile`` or ``HeifImage`` class to get images from.
        :param load_one: should be only one frame loaded. Default=``False``
        :param ignore_primary: force ``PrimaryImage=False`` flag to all added images."""

        if isinstance(heif_image, HeifFile):
            heif_images = list(heif_image)
        else:
            heif_images = [heif_image]
        for image in heif_images:
            additional_info = image.info.copy()
            if ignore_primary:
                additional_info["primary"] = False
            added_image = self.add_frombytes(
                image.mode,
                image.size,
                image.data,
                stride=image.stride,
                add_info={**additional_info},
                for_encoding=kwargs.get("for_encoding", False),
            )
            added_image.copy_thumbnails(image.thumbnails)
            if load_one:
                break
        return self

    def save(self, fp, **kwargs) -> None:
        """Saves image under the given fp.

        Keyword options can be used to provide additional instructions to the writer.
        If a writer does not recognise an option, it is silently ignored.

        Supported options:
            ``save_all`` - boolean. Should all images from ``HeiFile`` be saved.
            (default = ``True``)

            ``append_images`` - do the same as in Pillow.
            Accept ``HeifFile``, ``HeifImage`` and ``PIL.Image``

            .. note:: Appended images always will have ``info["primary"]=False``

            ``quality`` - see :py:attr:`~pillow_heif._options.PyLibHeifOptions.quality`

            ``enc_params`` - dictionary with key:value to pass to :ref:`x265 <hevc-encoder>` encoder.

            ``exif`` - override primary image's EXIF with specified. Accept ``None`` or ``bytes``.

            ``xmp`` - override primary image's XMP with specified. Accept ``None`` or ``bytes``.

            ``primary_index`` - ignore ``info["primary"]`` and set `PrimaryImage` by index.

            ``chroma`` - custom subsampling value. Possible values: ``444``, ``422`` or ``420`` (``x265`` default)
            Many software does not supports ``444`` chroma for HEIF.

        :param fp: A filename (string), pathlib.Path object or file object.

        :returns: None
        :raises: :py:exc:`~pillow_heif.HeifError` or :py:exc:`ValueError`"""

        compression = HeifCompressionFormat.HEVC
        if not have_encoder_for_format(compression):
            raise HeifError(code=HeifErrorCode.ENCODING_ERROR, subcode=5000, message="No encoder found.")
        images_to_save = self.__get_images_for_save(self.images, **kwargs)
        if not images_to_save:
            raise ValueError("Cannot write file with no images as HEIF.")
        primary_index = kwargs.pop("primary_index", None)
        if primary_index is None:
            primary_index = 0
            for i, img in enumerate(images_to_save):
                if img.info["primary"]:
                    primary_index = i
        elif primary_index == -1 or primary_index >= len(images_to_save):
            primary_index = len(images_to_save) - 1
        heif_ctx_write = LibHeifCtxWrite(compression_format=compression)
        enc_params = kwargs.get("enc_params", {})
        chroma = kwargs.get("chroma", None)
        if chroma:
            enc_params["chroma"] = chroma if isinstance(chroma, str) else str(chroma)
        heif_ctx_write.set_encoder_parameters(enc_params, kwargs.get("quality", options().quality))
        self._save(heif_ctx_write, images_to_save, primary_index, **kwargs)
        heif_ctx_write.write(fp)

    def __repr__(self):
        return f"<{self.__class__.__name__} with {len(self)} images: {[str(i) for i in self]}>"

    def __len__(self):
        return len(self.images)

    def __iter__(self):
        for _ in self.images:
            yield _

    def __getitem__(self, index):
        if index < 0 or index >= len(self.images):
            raise IndexError(f"invalid image index: {index}")
        return self.images[index]

    def __delitem__(self, key):
        if key < 0 or key >= len(self.images):
            raise IndexError(f"invalid image index: {key}")
        del self.images[key]

    def add_frombytes(self, mode: str, size: tuple, data, **kwargs):
        """Adds image from bytes to container.

        .. note:: Supports ``stride`` value if needed.

        :param mode: `BGR(A);16`, `L;16`, `I;16L`, `RGB(A);12`, `L;12`, `RGB(A);10`, `L;10`, `RGB(A)`, `BGR(A)`, `L`
        :param size: tuple with ``width`` and ``height`` of image.
        :param data: bytes object with raw image data.

        :returns: :py:class:`~pillow_heif.HeifImage` object that was appended to HeifFile."""

        for_encoding = kwargs.get("for_encoding", False)
        added_image = HeifImage(HeifCtxAsDict(mode, size, data, **kwargs), for_encoding=for_encoding)
        self.images.append(added_image)
        return added_image

    @property
    def __array_interface__(self):
        """Returns Primary Image as a numpy array."""

        return self.images[self.primary_index()].__array_interface__

    @staticmethod
    def _save(ctx: LibHeifCtxWrite, img_list: List[HeifImage], primary_index: int, **kwargs) -> None:
        enc_options = lib.heif_encoding_options_alloc()
        enc_options = ffi.gc(enc_options, lib.heif_encoding_options_free)
        for i, img in enumerate(img_list):
            set_color_profile(img.heif_img, img.info)
            p_img_handle = ffi.new("struct heif_image_handle **")
            error = lib.heif_context_encode_image(ctx.ctx, img.heif_img, ctx.encoder, enc_options, p_img_handle)
            check_libheif_error(error)
            new_img_handle = ffi.gc(p_img_handle[0], lib.heif_image_handle_release)
            exif = img.info["exif"]
            xmp = img.info["xmp"]
            if i == primary_index:
                if i:
                    lib.heif_context_set_primary_image(ctx.ctx, new_img_handle)
                if kwargs.get("exif", -1) != -1:
                    exif = kwargs["exif"]
                if kwargs.get("xmp", -1) != -1:
                    xmp = kwargs["xmp"]
            set_exif(ctx, new_img_handle, exif)
            set_xmp(ctx, new_img_handle, xmp)
            set_metadata(ctx, new_img_handle, img.info)
            for thumbnail in img.thumbnails:
                thumb_box = max(thumbnail.size)
                if max(img.size) > thumb_box > 3:
                    p_new_thumb_handle = ffi.new("struct heif_image_handle **")
                    error = lib.heif_context_encode_thumbnail(
                        ctx.ctx,
                        img.heif_img,
                        new_img_handle,
                        ctx.encoder,
                        enc_options,
                        thumb_box,
                        p_new_thumb_handle,
                    )
                    check_libheif_error(error)
                    if p_new_thumb_handle[0] != ffi.NULL:
                        lib.heif_image_handle_release(p_new_thumb_handle[0])

    @staticmethod
    def __get_images_for_save(images: List[HeifImage], **kwargs) -> List[HeifImage]:
        """Accepts list of Union[HeifFile, HeifImage, Image.Image] and returns List[HeifImage]"""

        images_to_save = images + list(kwargs.get("append_images", []))
        save_one = not kwargs.get("save_all", True)
        if save_one:
            images_to_save = images_to_save[:1]
        result = []
        for img in images_to_save:
            no_primary = not bool(img in images)
            if isinstance(img, Image.Image):
                heif_file = HeifFile().add_from_pillow(img, save_one, no_primary, for_encoding=True)
            elif kwargs.get("dont_copy", False) and bool(img in images):
                heif_file = [img] if isinstance(img, HeifImage) else img
            else:
                heif_file = HeifFile().add_from_heif(img, save_one, no_primary, for_encoding=True)
            result += list(heif_file)
        return result


def check_heif(fp):
    """Wrapper around `libheif.heif_check_filetype` function.

    .. note:: If `fp` contains less 12 bytes, then always return `HeifFiletype.NO`

    :param fp: See parameter ``fp`` in :func:`is_supported`

    :returns: Value from :py:class:`~pillow_heif.HeifFiletype` enumeration."""

    magic = _get_bytes(fp, 16)
    return HeifFiletype.NO if len(magic) < 12 else lib.heif_check_filetype(magic, len(magic))


def is_supported(fp) -> bool:
    """Checks if the given `fp` object contains a supported file type,
    by calling :py:func:`~pillow_heif.check_heif` function.

    Look at :py:attr:`~pillow_heif._options.PyLibHeifOptions.strict` property for additional info.

    :param fp: A filename (string), pathlib.Path object or a file object.
        The file object must implement ``file.read``,
        ``file.seek``, and ``file.tell`` methods,
        and be opened in binary mode.

    :returns: A boolean indicating if object can be opened."""

    magic = _get_bytes(fp, 16)
    heif_filetype = check_heif(magic)
    if heif_filetype == HeifFiletype.NO or (not options().avif and magic[8:12] in (b"avif", b"avis")):
        return False
    if heif_filetype in (HeifFiletype.YES_SUPPORTED, HeifFiletype.MAYBE):
        return True
    return not options().strict


def open_heif(fp, convert_hdr_to_8bit=True) -> HeifFile:
    """Opens the given HEIF image file.

    :param fp: See parameter ``fp`` in :func:`is_supported`
    :param convert_hdr_to_8bit: Boolean indicating should 10 bit or 12 bit images
        be converted to 8 bit images during loading.

    :returns: An :py:class:`~pillow_heif.HeifFile` object.
    :exception HeifError: If file is corrupted or is not in Heif format."""

    heif_ctx = LibHeifCtx(fp, convert_hdr_to_8bit)
    return HeifFile(heif_ctx, heif_ctx.get_top_images_ids(), heif_ctx.get_main_img_id())


def read_heif(fp, convert_hdr_to_8bit=True) -> HeifFile:
    """Opens the given HEIF image file and decodes all images.

    .. note:: In most cases it is better to call :py:meth:`~pillow_heif.open_heif`, and
        let images decoded automatically only when needed.

    :param fp: See parameter ``fp`` in :func:`is_supported`
    :param convert_hdr_to_8bit: Boolean indicating should 10 bit or 12 bit images
        be converted to 8 bit images during loading.

    :returns: An :py:class:`~pillow_heif.HeifFile` object.
    :exception HeifError: If file is corrupted or is not in Heif format."""

    heif_file = open_heif(fp, convert_hdr_to_8bit)
    heif_file.load(everything=True)
    return heif_file


def from_pillow(pil_image: Image.Image, load_one: bool = False, ignore_primary=True) -> HeifFile:
    """Creates :py:class:`~pillow_heif.HeifFile` from a Pillow Image.

    :param pil_image: Pillow :external:py:class:`~PIL.Image.Image` class.
    :param load_one: If ``True``, then all frames will be loaded.
    :param ignore_primary: force ``PrimaryImage=False`` flag to all added images.

    :returns: An :py:class:`~pillow_heif.HeifFile` object."""

    return HeifFile().add_from_pillow(pil_image, load_one, ignore_primary)


def from_bytes(mode: str, size: tuple, data, **kwargs) -> HeifFile:
    """Creates :py:class:`~pillow_heif.HeifFile` from bytes.

    .. note:: Supports ``stride`` value if needed.

    :param mode: `BGR(A);16`, `L;16`, `I;16L`, `RGB(A);12`, `L;12`, `RGB(A);10`, `L;10`, `RGB(A)`, `BGR(A)`, `L`
    :param size: tuple with ``width`` and ``height`` of image.
    :param data: bytes object with raw image data.

    :returns: An :py:class:`~pillow_heif.HeifFile` object."""

    _ = HeifFile()
    _.add_frombytes(mode, size, data, **kwargs)
    return _


# --------------------------------------------------------------------
# DEPRECATED FUNCTIONS.
# pylint: disable=unused-argument
# pylint: disable=redefined-builtin


def check(fp):
    warn("Function `check` is deprecated, use `check_heif` instead.", DeprecationWarning)
    return check_heif(fp)  # pragma: no cover


def open(fp, *, apply_transformations=True, convert_hdr_to_8bit=True):  # noqa
    warn("Function `open` is deprecated and will be removed, use `open_heif` instead.", DeprecationWarning)
    return open_heif(fp, convert_hdr_to_8bit=convert_hdr_to_8bit)  # pragma: no cover


def read(fp, *, apply_transformations=True, convert_hdr_to_8bit=True):  # noqa
    warn("Function `read` is deprecated and will be removed, use `open_heif` instead.", DeprecationWarning)
    return open_heif(fp, convert_hdr_to_8bit=convert_hdr_to_8bit)  # pragma: no cover



from io import BytesIO

import helpers
import pytest
from packaging.version import parse as parse_version
from PIL import Image

import pillow_heif

np = pytest.importorskip("numpy", reason="NumPy not installed")

pillow_heif.register_heif_opener()
if not helpers.hevc_enc():
    pytest.skip(reason="Requires HEIF encoder.", allow_module_level=True)

# Creating HEIF file in memory with 1 image and 3 thumbnails.
im_pillow = Image.effect_mandelbrot((512, 512), (-3, -2.5, 2, 2.5), 100)
im_heif = pillow_heif.from_pillow(im_pillow)
pillow_heif.add_thumbnails(im_heif, boxes=[192, 128, 64])
heif_buf = BytesIO()
im_heif.save(heif_buf)


@pytest.mark.parametrize("im_size", ((256, 128), (127, 64), (63, 64), (31, 32), (20, 100), (14, 16), (11, 16)))
@pytest.mark.parametrize("mode", ("L", "RGB", "RGBA", "I;16"))
def test_numpy_array(im_size, mode):
    im = im_pillow.convert(mode=mode)
    heif_file = pillow_heif.from_pillow(im)
    pil_array = np.asarray(im)
    heif_array = np.asarray(heif_file[0])
    assert np.array_equal(pil_array, heif_array)
    assert np.array_equal(pil_array, np.asarray(heif_file))


@pytest.mark.skipif(parse_version(np.__version__) < parse_version("1.23.0"), reason="Requires numpy >= 1.23")
def test_numpy_array_pillow_thumbnail():
    pil_img = Image.open(heif_buf)
    thumbnail = pillow_heif.thumbnail(pil_img)
    assert thumbnail.size != pil_img.size
    np.asarray(thumbnail)
    for i in range(3):
        np.asarray(pil_img.info["thumbnails"][i])
    pil_img.load()
    for i in range(3):
        with pytest.raises(ValueError):
            np.asarray(pil_img.info["thumbnails"][i])


def test_numpy_array_heif_thumbnail():
    heif_file = pillow_heif.open_heif(heif_buf)
    thumbnail = pillow_heif.thumbnail(heif_file)
    assert thumbnail.size != heif_file.size
    np.asarray(thumbnail)
    for i in range(3):
        np.asarray(heif_file.thumbnails[i])



"""
Undocumented private functions for other code to look better.
"""

from math import ceil
from typing import Union

from _pillow_heif_cffi import ffi, lib
from PIL import __version__ as pil_version

from ._libheif_ctx import LibHeifCtxWrite
from .constants import HeifChroma, HeifColorProfileType, HeifColorspace
from .error import check_libheif_error

FFI_DRY_ALLOC = ffi.new_allocator(should_clear_after_alloc=False)


# from dataclasses import dataclass
# @dataclass                # Avalaible from Python 3.7
class HeifCtxAsDict:  # noqa # pylint: disable=too-few-public-methods
    """Representation of one image"""

    def __init__(self, mode: str, size: tuple, data, **kwargs):
        stride = kwargs.get("stride", None)
        if stride is None:
            stride = get_pure_stride(mode, size[0])
        self.mode = mode
        self.size = size
        self.data = data
        self.stride = stride
        self.additional_info = kwargs.get("add_info", {})


def get_pure_stride(mode: str, width: int):
    return width * MODE_INFO[mode][0] * ceil(MODE_INFO[mode][1] / 8)


MODE_CONVERT = {
    # source_mode: {target_mode: convert_function,}
    "BGRA;16": {"RGBA;10": lib.convert_bgra16_to_rgba10, "RGBA;12": lib.convert_bgra16_to_rgba12},
    "BGR;16": {"RGB;10": lib.convert_bgr16_to_rgb10, "RGB;12": lib.convert_bgr16_to_rgb12},
    "RGBA;16": {"RGBA;10": lib.convert_rgba16_to_rgba10, "RGBA;12": lib.convert_rgba16_to_rgba12},
    "RGB;16": {"RGB;10": lib.convert_rgb16_to_rgb10, "RGB;12": lib.convert_rgb16_to_rgb12},
    "L;16": {"L;10": lib.convert_i16_to_i10, "L;12": lib.convert_i16_to_i12},
    "I;16": {"L;10": lib.convert_i16_to_i10, "L;12": lib.convert_i16_to_i12},
    "I;16L": {"L;10": lib.convert_i16_to_i10, "L;12": lib.convert_i16_to_i12},
    "RGBA;12": {"RGBA;16": lib.convert_rgba12_to_rgba16, "BGRA;16": lib.convert_rgba12_to_bgra16},
    "RGB;12": {"RGB;16": lib.convert_rgb12_to_rgb16, "BGR;16": lib.convert_rgb12_to_bgr16},
    "RGBA;10": {"RGBA;16": lib.convert_rgba10_to_rgba16, "BGRA;16": lib.convert_rgba10_to_bgra16},
    "RGB;10": {"RGB;16": lib.convert_rgb10_to_rgb16, "BGR;16": lib.convert_rgb10_to_bgr16},
    "BGRA": {"RGBA": lib.convert_bgra_rgba},
    "BGR": {"RGB": lib.convert_bgr_rgb},
    "RGBA": {
        "BGRA": lib.convert_bgra_rgba,
        "RGBA;16": lib.convert_rgba_to_rgba16,
        "BGRA;16": lib.convert_rgba_to_bgra16,
    },
    "RGB": {"BGR": lib.convert_bgr_rgb, "RGB;16": lib.convert_rgb_to_rgb16, "BGR;16": lib.convert_rgb_to_bgr16},
}


MODE_INFO = {
    # name -> [channels, bits per pixel channel, colorspace, chroma, (mode_for_saving 10,12 bit), numpy_typestr]
    "BGRA;16": (4, 16, HeifColorspace.RGB, HeifChroma.INTERLEAVED_RRGGBBAA_LE, ("RGBA;10", "RGBA;12"), "<u2"),
    "BGR;16": (3, 16, HeifColorspace.RGB, HeifChroma.INTERLEAVED_RRGGBB_LE, ("RGB;10", "RGB;12"), "<u2"),
    "RGBA;16": (4, 16, HeifColorspace.RGB, HeifChroma.INTERLEAVED_RRGGBBAA_LE, ("RGBA;10", "RGBA;12"), "<u2"),
    "RGB;16": (3, 16, HeifColorspace.RGB, HeifChroma.INTERLEAVED_RRGGBB_LE, ("RGB;10", "RGB;12"), "<u2"),
    "L;16": (1, 16, HeifColorspace.MONOCHROME, HeifChroma.MONOCHROME, ("L;10", "L;12"), "<u2"),
    "I;16": (1, 16, HeifColorspace.MONOCHROME, HeifChroma.MONOCHROME, ("L;10", "L;12"), "<u2"),
    "I;16L": (1, 16, HeifColorspace.MONOCHROME, HeifChroma.MONOCHROME, ("L;10", "L;12"), "<u2"),
    "RGBA;12": (4, 12, HeifColorspace.RGB, HeifChroma.INTERLEAVED_RRGGBBAA_LE, None, "<u2"),
    "RGB;12": (3, 12, HeifColorspace.RGB, HeifChroma.INTERLEAVED_RRGGBB_LE, None, "<u2"),
    "RGBA;12B": (4, 12, HeifColorspace.RGB, HeifChroma.INTERLEAVED_RRGGBBAA_BE, None, ">u2"),
    "RGB;12B": (3, 12, HeifColorspace.RGB, HeifChroma.INTERLEAVED_RRGGBB_BE, None, ">u2"),
    "L;12": (1, 12, HeifColorspace.MONOCHROME, HeifChroma.MONOCHROME, None, "<u2"),
    "RGBA;10": (4, 10, HeifColorspace.RGB, HeifChroma.INTERLEAVED_RRGGBBAA_LE, None, "<u2"),
    "RGB;10": (3, 10, HeifColorspace.RGB, HeifChroma.INTERLEAVED_RRGGBB_LE, None, "<u2"),
    "RGBA;10B": (4, 10, HeifColorspace.RGB, HeifChroma.INTERLEAVED_RRGGBBAA_BE, None, ">u2"),
    "RGB;10B": (3, 10, HeifColorspace.RGB, HeifChroma.INTERLEAVED_RRGGBB_BE, None, ">u2"),
    "L;10": (1, 10, HeifColorspace.MONOCHROME, HeifChroma.MONOCHROME, None, "<u2"),
    "RGBA": (4, 8, HeifColorspace.RGB, HeifChroma.INTERLEAVED_RGBA, None, "|u1"),
    "RGB": (3, 8, HeifColorspace.RGB, HeifChroma.INTERLEAVED_RGB, None, "|u1"),
    "BGRA": (4, 8, HeifColorspace.RGB, HeifChroma.INTERLEAVED_RGBA, "RGBA", "|u1"),
    "BGR": (3, 8, HeifColorspace.RGB, HeifChroma.INTERLEAVED_RGB, "RGB", "|u1"),
    "L": (1, 8, HeifColorspace.MONOCHROME, HeifChroma.MONOCHROME, None, "|u1"),
    "": (0, 0, HeifColorspace.UNDEFINED, HeifChroma.UNDEFINED, None, ""),
}


def read_color_profile(handle) -> dict:
    profile_type = lib.heif_image_handle_get_color_profile_type(handle)
    if profile_type == HeifColorProfileType.NOT_PRESENT:
        return {}
    if profile_type == HeifColorProfileType.NCLX:
        _type = "nclx"
        pp_data = ffi.new("struct heif_color_profile_nclx **")
        data_length = ffi.sizeof("struct heif_color_profile_nclx")
        error = lib.heif_image_handle_get_nclx_color_profile(handle, pp_data)
        p_data = pp_data[0]
        ffi.release(pp_data)
    else:
        _type = "prof" if profile_type == HeifColorProfileType.PROF else "rICC"
        data_length = lib.heif_image_handle_get_raw_color_profile_size(handle)
        if data_length == 0:
            return {"type": _type, "data": b""}
        p_data = ffi.new("char[]", data_length)
        error = lib.heif_image_handle_get_raw_color_profile(handle, p_data)
    check_libheif_error(error)
    data_buffer = ffi.buffer(p_data, data_length)
    return {"type": _type, "data": bytes(data_buffer)}


def set_color_profile(heif_img, info: dict) -> None:
    __icc_profile = info.get("icc_profile", None)
    if __icc_profile is not None:
        _prof_type = info.get("icc_profile_type", "prof").encode("ascii")
        error = lib.heif_image_set_raw_color_profile(
            heif_img, _prof_type, info["icc_profile"], len(info["icc_profile"])
        )
        check_libheif_error(error)
    elif info.get("nclx_profile", None):
        error = lib.heif_image_set_nclx_color_profile(
            heif_img,
            ffi.cast("const struct heif_color_profile_nclx*", ffi.from_buffer(info["nclx_profile"])),
        )
        check_libheif_error(error)


def retrieve_exif(metadata: list):
    _result = None
    _purge = []
    for i, md_block in enumerate(metadata):
        if md_block["type"] == "Exif":
            _purge.append(i)
            if not _result and md_block["data"] and md_block["data"][0:4] == b"Exif":
                _result = md_block["data"]
    for i in reversed(_purge):
        del metadata[i]
    return _result


def set_exif(ctx: LibHeifCtxWrite, heif_img_handle, exif: Union[bytes, None]) -> None:
    if exif is not None:
        error = lib.heif_context_add_exif_metadata(ctx.ctx, heif_img_handle, exif, len(exif))
        check_libheif_error(error)


def retrieve_xmp(metadata: list):
    _result = None
    _purge = []
    for i, md_block in enumerate(metadata):
        if md_block["type"] == "mime":
            _purge.append(i)
            if not _result:
                _result = md_block["data"]
    for i in reversed(_purge):
        del metadata[i]
    return _result


def set_xmp(ctx: LibHeifCtxWrite, heif_img_handle, xmp: Union[bytes, None]) -> None:
    if xmp is not None:
        error = lib.heif_context_add_XMP_metadata(ctx.ctx, heif_img_handle, xmp, len(xmp))
        check_libheif_error(error)


def read_metadata(handle) -> list:
    block_count = lib.heif_image_handle_get_number_of_metadata_blocks(handle, ffi.NULL)
    if block_count == 0:
        return []
    metadata = []
    blocks_ids = ffi.new("heif_item_id[]", block_count)
    lib.heif_image_handle_get_list_of_metadata_block_IDs(handle, ffi.NULL, blocks_ids, block_count)
    for block_id in blocks_ids:
        metadata_type = lib.heif_image_handle_get_metadata_type(handle, block_id)
        metadata_type = ffi.string(metadata_type).decode("utf-8")
        content_type = ffi.string(lib.heif_image_handle_get_metadata_content_type(handle, block_id))
        data_length = lib.heif_image_handle_get_metadata_size(handle, block_id)
        if data_length > 0:
            p_data = ffi.new("char[]", data_length)
            error = lib.heif_image_handle_get_metadata(handle, block_id, p_data)
            check_libheif_error(error)
            data_buffer = ffi.buffer(p_data, data_length)
            data = bytes(data_buffer)
            if metadata_type == "Exif":
                data = data[4:]  # skip TIFF header, first 4 bytes
            metadata.append({"type": metadata_type, "data": data, "content_type": content_type})
    return metadata


def set_metadata(ctx: LibHeifCtxWrite, heif_img_handle, info: dict) -> None:
    for metadata in info["metadata"]:
        error = lib.heif_context_add_generic_metadata(
            ctx.ctx,
            heif_img_handle,
            metadata["data"],
            len(metadata["data"]),
            metadata["type"].encode("utf-8"),
            metadata["content_type"],
        )
        check_libheif_error(error)


def exif_from_pillow(additional_info: dict, frame) -> None:
    if "exif" not in additional_info:
        if hasattr(frame, "getexif"):
            if pil_version[:2] not in ("6.", "7.", "8.") and pil_version[:4] not in ("9.0.", "9.1."):
                exif = frame.getexif()
                if exif:
                    additional_info["exif"] = exif.tobytes()


def xmp_from_pillow(additional_info: dict, frame) -> None:
    if "xmp" not in additional_info:
        if "XML:com.adobe.xmp" in frame.info:  # PNG
            additional_info["xmp"] = frame.info["XML:com.adobe.xmp"]
        elif hasattr(frame, "tag_v2"):  # TIFF
            if 700 in frame.tag_v2:
                additional_info["xmp"] = frame.tag_v2[700]
        elif hasattr(frame, "applist"):  # JPEG
            for segment, content in frame.applist:
                if segment == "APP1":
                    marker, xmp_tags = content.rsplit(b"\x00", 1)
                    if marker == b"http://ns.adobe.com/xap/1.0/":
                        additional_info["xmp"] = xmp_tags
                        break
    if "xmp" in additional_info and isinstance(additional_info["xmp"], str):
        additional_info["xmp"] = additional_info["xmp"].encode("utf-8")



"""
Plugin for Pillow library.
"""

from typing import Any

from PIL import Image, ImageFile

from ._lib_info import have_encoder_for_format
from ._options import options
from .constants import HeifCompressionFormat, HeifErrorCode
from .error import HeifError
from .heif import HeifFile, is_supported, open_heif
from .misc import getxmp, set_orientation


class _LibHeifImageFile(ImageFile.ImageFile):
    """Base class with all functionality for ``HeifImageFile`` and ``AvifImageFile`` classes."""

    heif_file: Any
    _close_exclusive_fp_after_loading = False

    def __init__(self, *args, **kwargs):
        self.__frame = 0
        self.heif_file = None
        super().__init__(*args, **kwargs)

    def _open(self):
        try:
            heif_file = open_heif(self.fp)
        except HeifError as exception:
            raise SyntaxError(str(exception)) from None
        self.custom_mimetype = heif_file.mimetype
        self.heif_file = heif_file
        self.__frame = heif_file.primary_index()
        self._init_from_heif_file(self.__frame)
        self.tile = []

    def load(self):
        if self.heif_file:
            frame_heif = self.heif_file[self.tell()]
            self.load_prepare()
            try:
                self.frombytes(frame_heif.data, "raw", (frame_heif.mode, frame_heif.stride))
            except HeifError as exc:
                truncated = exc.code == HeifErrorCode.DECODER_PLUGIN_ERROR and exc.subcode == 100
                if not truncated or not ImageFile.LOAD_TRUNCATED_IMAGES:
                    raise
            if not self.is_animated:
                self.info["thumbnails"] = [i.clone_nd() for i in self.info["thumbnails"]]
                self.heif_file = None
                self._close_exclusive_fp_after_loading = True
                if self.fp and getattr(self, "_exclusive_fp", False) and hasattr(self.fp, "close"):
                    self.fp.close()
                self.fp = None
        return super().load()

    def getxmp(self) -> dict:
        """Returns a dictionary containing the XMP tags. Requires ``defusedxml`` to be installed.

        :returns: XMP tags in a dictionary."""

        return getxmp(self.info["xmp"])

    def seek(self, frame):
        if not self._seek_check(frame):
            return
        self.__frame = frame
        self._init_from_heif_file(frame)
        _exif = getattr(self, "_exif", None)  # Pillow 9.2+ do no reload exif between frames.
        if _exif is not None:
            if getattr(_exif, "_loaded", None):
                _exif._loaded = False  # pylint: disable=protected-access

    def tell(self):
        return self.__frame

    def verify(self) -> None:
        pass

    @property
    def n_frames(self) -> int:
        """Return the number of available frames.

        :returns: Frame number, starting with 0."""

        return len(self.heif_file) if self.heif_file else 1

    @property
    def is_animated(self) -> bool:
        """Return ``True`` if this image has more then one frame, or ``False`` otherwise."""

        return self.n_frames > 1

    def _seek_check(self, frame):
        if frame < 0 or frame >= self.n_frames:
            raise EOFError("attempt to seek outside sequence")
        return self.tell() != frame

    def _init_from_heif_file(self, img_index: int) -> None:
        self._size = self.heif_file[img_index].size
        self.mode = self.heif_file[img_index].mode
        self.info = self.heif_file[img_index].info
        self.info["thumbnails"] = self.heif_file[img_index].thumbnails
        self.info["original_orientation"] = set_orientation(self.info)


class HeifImageFile(_LibHeifImageFile):
    """Pillow plugin class type for a HEIF image format."""

    format = "HEIF"
    format_description = "HEIF container"


def _save_heif(im, fp, _filename):
    heif_file = HeifFile().add_from_pillow(im, load_one=True, for_encoding=True)
    heif_file.save(fp, save_all=False, **im.encoderinfo, dont_copy=True)


def _save_all_heif(im, fp, _filename):
    heif_file = HeifFile().add_from_pillow(im, ignore_primary=False, for_encoding=True)
    heif_file.save(fp, save_all=True, **im.encoderinfo, dont_copy=True)


def register_heif_opener(**kwargs) -> None:
    """Registers a Pillow plugin for HEIF format.

    :param kwargs: dictionary with values to set in :py:class:`~pillow_heif._options.PyLibHeifOptions`
    """

    options().update(**kwargs)
    Image.register_open(HeifImageFile.format, HeifImageFile, is_supported)
    if have_encoder_for_format(HeifCompressionFormat.HEVC):
        Image.register_save(HeifImageFile.format, _save_heif)
        Image.register_save_all(HeifImageFile.format, _save_all_heif)
    extensions = [".heic", ".heics", ".heif", ".heifs", ".hif"]
    Image.register_mime(HeifImageFile.format, "image/heic")
    Image.register_mime(HeifImageFile.format, "image/heic-sequence")
    Image.register_mime(HeifImageFile.format, "image/heif")
    Image.register_mime(HeifImageFile.format, "image/heif-sequence")
    Image.register_extensions(HeifImageFile.format, extensions)



"""Functions for work with thumbnails."""

from typing import List, Union

from PIL import Image

from .as_opener import _LibHeifImageFile
from .heif import HeifFile, HeifImage, HeifThumbnail
from .private import HeifCtxAsDict


def thumbnail(im, min_box: int = 0):
    """Returns a thumbnail with minimum specified box size or an original if there is no a thumbnail.

    .. note:: If you use it for a Pillow image, it ``should not be loaded`` before call.

    :param im: :py:class:`HeifImage`, :py:class:`HeifFile` or :external:py:class:`PIL.Image.Image` class.
    :param min_box: optional integer with minimum size of thumbnail's box size.

    :returns: Class with the same type the input parameter ``im`` has."""

    if isinstance(im, HeifFile):
        im = im.images[im.primary_index()]
    thumb = None
    if isinstance(im, HeifImage):
        thumb = _choose_thumbnail(im.thumbnails, min_box)
        if thumb:
            thumb.info = im.info
    elif isinstance(im, _LibHeifImageFile):
        thumb = _choose_thumbnail(im.info.get("thumbnails", []), min_box)
        if thumb:
            thumb = thumb.to_pillow()
    return thumb if thumb else im


def add_thumbnails(im: Union[HeifFile, HeifImage, Image.Image], boxes: Union[List[int], int]) -> None:
    """Add thumbnail(s) to an image(s).

    .. note:: Method creates thumbnails without image data, they will be encoded during `save` operation.

    :param im: For ``HeifFile`` will add thumbnail(s) to all images in it.
        For ``HeifImage`` or a Pillow image it will add thumbnails only to specified image.
    :param boxes: int or list of ints determining size of thumbnail(s) to generate for image.

    :returns: None"""

    if isinstance(im, HeifFile):
        for i in im:
            _add_thumbnails(i, boxes)
    else:
        _add_thumbnails(im, boxes)


def _add_thumbnails(im: Union[HeifImage, Image.Image], boxes: Union[List[int], int]) -> None:
    if isinstance(boxes, list):
        boxes_list = boxes
    else:
        boxes_list = [boxes]
    for box in boxes_list:
        if box <= 3:
            continue
        if im.size[0] <= box and im.size[1] <= box:
            continue
        if im.size[0] > im.size[1]:
            thumb_height = int(im.size[1] * box / im.size[0])
            thumb_width = box
        else:
            thumb_width = int(im.size[0] * box / im.size[1])
            thumb_height = box
        thumb_height = thumb_height - 1 if (thumb_height & 1) else thumb_height
        thumb_width = thumb_width - 1 if (thumb_width & 1) else thumb_width
        thumbnails = im.thumbnails if isinstance(im, HeifImage) else im.info.get("thumbnails", [])
        if max((thumb_height, thumb_width)) in [max(i.size) for i in thumbnails]:
            continue
        __heif_ctx = HeifCtxAsDict(im.mode, (thumb_width, thumb_height), None, stride=0)
        if isinstance(im, HeifImage):
            im.thumbnails.append(HeifThumbnail(__heif_ctx, im))
        else:
            if im.info.get("thumbnails", None) is None:
                im.info["thumbnails"] = [HeifThumbnail(__heif_ctx, __heif_ctx)]
            else:
                im.info["thumbnails"].append(HeifThumbnail(__heif_ctx, __heif_ctx))


def _choose_thumbnail(thumbnails: list, min_box: int):
    for thumb in thumbnails:
        if thumb.size[0] >= min_box or thumb.size[1] >= min_box:
            if thumb.data:
                return thumb
    return None



import os
from io import BytesIO
from pathlib import Path

import helpers
import pytest
from packaging.version import parse as parse_version
from PIL import Image
from PIL import __version__ as pil_version
from PIL import features

import pillow_heif

os.chdir(os.path.dirname(os.path.abspath(__file__)))
pillow_heif.register_heif_opener()


@pytest.mark.skipif(not features.check("webp"), reason="Requires WEBP support.")
@pytest.mark.skipif(not helpers.hevc_enc(), reason="Requires HEIF encoder.")
@pytest.mark.skipif(parse_version(pil_version) < parse_version("8.3.0"), reason="Requires Pillow >= 8.3")
@pytest.mark.parametrize(
    "img_path",
    (
        "images/non_heif/xmp.png",
        "images/non_heif/xmp.jpeg",
        "images/non_heif/xmp.tiff",
        "images/non_heif/xmp.webp",
    ),
)
def test_xmp_from_pillow(img_path):
    im = Image.open(Path(img_path))
    xmp = im.getxmp() if hasattr(im, "getxmp") else pillow_heif.getxmp(im.info["xmp"])  # noqa
    assert xmp["xmpmeta"]["RDF"]["Description"]["subject"]["Bag"]["li"] == "TestSubject"
    out_im_heif = BytesIO()
    im.save(out_im_heif, format="HEIF")
    im_heif = Image.open(out_im_heif)
    assert im_heif.info["xmp"]
    assert isinstance(im_heif.info["xmp"], bytes)
    xmp_heif = im_heif.getxmp()  # noqa
    assert xmp_heif["xmpmeta"]["RDF"]["Description"]["subject"]["Bag"]["li"] == "TestSubject"


@pytest.mark.skipif(not helpers.hevc_enc(), reason="Requires HEIF encoder.")
def test_pillow_xmp_add_remove():
    xmp_data = b"<xmp_data>"
    out_heif_no_xmp = BytesIO()
    out_heif = BytesIO()
    im = Image.new("RGB", (15, 15), 0)
    # filling `xmp` during save.
    im.save(out_heif, format="HEIF", xmp=xmp_data)
    assert "xmp" not in im.info or im.info["xmp"] is None
    im_heif = Image.open(out_heif)
    assert im_heif.info["xmp"]
    # filling `info["xmp"]` before save.
    im.info["xmp"] = xmp_data
    im.save(out_heif, format="HEIF")
    im_heif = Image.open(out_heif)
    assert im_heif.info["xmp"]
    # setting `xmp` to `None` during save.
    im_heif.save(out_heif_no_xmp, format="HEIF", xmp=None)
    assert im_heif.info["xmp"]
    im_heif_no_xmp = Image.open(out_heif_no_xmp)
    assert "xmp" not in im_heif_no_xmp.info or im_heif_no_xmp.info["xmp"] is None
    # filling `info["xmp"]` with `None` before save.
    im_heif.info["xmp"] = None
    im_heif.save(out_heif_no_xmp, format="HEIF")
    im_heif_no_xmp = Image.open(out_heif_no_xmp)
    assert "xmp" not in im_heif_no_xmp.info or im_heif_no_xmp.info["xmp"] is None
    # removing `info["xmp"]` before save.
    im_heif.info.pop("xmp")
    im_heif.save(out_heif_no_xmp, format="HEIF")
    im_heif_no_xmp = Image.open(out_heif_no_xmp)
    assert "xmp" not in im_heif_no_xmp.info or im_heif_no_xmp.info["xmp"] is None


@pytest.mark.skipif(not helpers.hevc_enc(), reason="Requires HEIF encoder.")
def test_heif_xmp_add_remove():
    xmp_data = b"<xmp_data>"
    out_heif_no_xmp = BytesIO()
    out_heif = BytesIO()
    # test filling `xmp` during save.
    im_heif = pillow_heif.from_pillow(Image.new("RGB", (15, 15), 0))
    im_heif.save(out_heif, xmp=xmp_data)
    assert "xmp" not in im_heif.info or im_heif.info["xmp"] is None
    im_heif = pillow_heif.open_heif(out_heif)
    assert im_heif.info["xmp"]
    # test filling `info["xmp"]` before save.
    im_heif = pillow_heif.from_pillow(Image.new("RGB", (15, 15), 0))
    im_heif.info["xmp"] = xmp_data
    im_heif.save(out_heif)
    im_heif = pillow_heif.open_heif(out_heif)
    assert im_heif.info["xmp"]
    # setting `xmp` to `None` during save.
    im_heif.save(out_heif_no_xmp, xmp=None)
    assert im_heif.info["xmp"]
    im_heif_no_xmp = pillow_heif.open_heif(out_heif_no_xmp)
    assert "xmp" not in im_heif_no_xmp.info or im_heif_no_xmp.info["xmp"] is None
    # filling `info["xmp"]` with `None` before save.
    im_heif.info["xmp"] = None
    im_heif.save(out_heif_no_xmp)
    im_heif_no_xmp = pillow_heif.open_heif(out_heif_no_xmp)
    assert "xmp" not in im_heif_no_xmp.info or im_heif_no_xmp.info["xmp"] is None
    # removing `info["xmp"]` before save.
    im_heif.info.pop("xmp")
    im_heif.save(out_heif_no_xmp)
    im_heif_no_xmp = pillow_heif.open_heif(out_heif_no_xmp)
    assert "xmp" not in im_heif_no_xmp.info or im_heif_no_xmp.info["xmp"] is None



"""
Options to change pillow_heif's runtime behaviour.
"""
from typing import Union
from warnings import warn

from ._lib_info import have_decoder_for_format, have_encoder_for_format
from .constants import HeifCompressionFormat


class PyLibHeifOptions:
    """Pillow-Heif runtime properties."""

    def __init__(self):
        self._hevc_enc = have_encoder_for_format(HeifCompressionFormat.HEVC)
        self._avif_dec = have_decoder_for_format(HeifCompressionFormat.AV1)
        self._cfg = {}
        self.reset()

    @property
    def hevc_enc(self) -> bool:
        """Read only property showing if library was build with h265 HEVC encoder."""

        warn("Started from `0.6.0` version this property will be deprecated.", DeprecationWarning)
        return self._hevc_enc

    @property
    def avif(self) -> bool:
        """Enable or disable x264(AVIF) file read support. Default=True(if build include libaom)"""

        warn("Started from `0.6.0` version this property will be deprecated.", DeprecationWarning)
        return self._cfg["avif"]

    @avif.setter
    def avif(self, value: bool):
        warn("Started from `0.6.0` version this property will be deprecated.", DeprecationWarning)
        self._cfg["avif"] = value if self._avif_dec else False

    @property
    def strict(self) -> bool:
        """Indicates should be or not partially supported files be marked as supported.
        Affects on return result of :py:func:`~pillow_heif.is_supported` function.

        Default = ``False``"""

        return self._cfg["strict"]

    @strict.setter
    def strict(self, value: bool):
        self._cfg["strict"] = value

    @property
    def thumbnails(self) -> bool:
        """Property to enable or disable HEIF thumbnails functionality.

        Default = ``True``"""

        return self._cfg["thumbnails"]

    @thumbnails.setter
    def thumbnails(self, value: bool):
        self._cfg["thumbnails"] = value

    @property
    def quality(self) -> Union[int, None]:
        """Get or set default encoding quality.

        .. note:: Quality specified during :py:meth:`~pillow_heif.HeifFile.save`
            has higher priority then this.

        Possible values: None, -1, range(0-100). Default=None
            Set -1 for lossless quality or from 0 to 100, where 0 is lowest and 100 is highest.

        .. note:: Also for lossless encoding you should specify ``chroma=444`` during save."""

        return self._cfg["quality"]

    @quality.setter
    def quality(self, value):
        self._cfg["quality"] = value

    @property
    def ctx_in_memory(self) -> bool:
        """Should files be read to memory fully.

        As Python is much slower than ``C++``, we read a file to memory and let ``libheif`` manage reads.
        You can look at source of class ``LibHeifCtx`` in ``_libheif_ctx.py`` file.
        This will be the default behaviour, until maybe someday written on ``C`` reader module will appear.

        Default = ``True``"""

        return self._cfg["ctx_in_memory"]

    @ctx_in_memory.setter
    def ctx_in_memory(self, value):
        self._cfg["ctx_in_memory"] = value

    @property
    def save_to_12bit(self) -> bool:
        """Should 16 bit images be saved to 12 bit instead of 10. Default = ``False``"""

        return self._cfg["save_to_12bit"]

    @save_to_12bit.setter
    def save_to_12bit(self, value):
        self._cfg["save_to_12bit"] = value

    def update(self, **kwargs) -> None:
        """Method for at once update multiply values in config."""

        _keys = kwargs.keys()
        for k in ("avif", "strict", "thumbnails", "quality", "ctx_in_memory", "save_to_12bit"):
            if k in _keys:
                setattr(self, k, kwargs[k])

    def reset(self) -> None:
        """Use this for reset config values to their defaults."""

        self._cfg["avif"] = self._avif_dec
        self._cfg["strict"] = False
        self._cfg["thumbnails"] = True
        self._cfg["quality"] = None
        self._cfg["ctx_in_memory"] = True
        self._cfg["save_to_12bit"] = False


CFG_OPTIONS: PyLibHeifOptions = PyLibHeifOptions()


def options() -> PyLibHeifOptions:
    """Wrapper function to return runtime options variable.

    :returns: A :py:class:`~pillow_heif._options.PyLibHeifOptions` class."""

    return CFG_OPTIONS



"""
Different miscellaneous helper functions.

Mostly for internal use, so prototypes can change between versions.
"""

import builtins
import pathlib
import re
from struct import pack, unpack
from typing import Union
from warnings import warn

from _pillow_heif_cffi import ffi, lib

try:
    from defusedxml import ElementTree
except ImportError:  # pragma: no cover
    ElementTree = None  # pragma: no cover


def set_orientation(info: dict, orientation: int = 1) -> Union[int, None]:
    """Sets orientation in ``EXIF`` to ``1`` by default if any orientation present.
    Removes ``XMP`` orientation tag if it is present.
    In Pillow plugin mode it called automatically for images.
    When ``pillow_heif`` used in ``standalone`` mode, if you wish you can call it manually.

    .. note:: If there is no orientation tag, this function will not add it and do nothing.

        If both XMP and EXIF orientation tags present, EXIF orientation tag will be returned,
        but both tags will be removed.

    :param info: `info` dictionary from `~PIL.Image.Image` or `~pillow_heif.HeifImage`.
    :param orientation: int value of EXIF or XMP orientation tag.
    :returns: Original orientation or None if it is absent."""

    original_orientation = None
    if info.get("exif", None):
        tif_tag = info["exif"][6:]
        endian_mark = "<" if tif_tag[0:2] == b"\x49\x49" else ">"
        pointer = unpack(endian_mark + "L", tif_tag[4:8])[0]
        tag_count = unpack(endian_mark + "H", tif_tag[pointer : pointer + 2])[0]
        offset = pointer + 2
        for tag_n in range(tag_count):
            pointer = offset + 12 * tag_n
            if unpack(endian_mark + "H", tif_tag[pointer : pointer + 2])[0] != 274:
                continue
            value = tif_tag[pointer + 8 : pointer + 12]
            _original_orientation = unpack(endian_mark + "H", value[0:2])[0]
            if _original_orientation != 1:
                original_orientation = _original_orientation
                p_value = 6 + pointer + 8
                new_orientation = pack(endian_mark + "H", orientation)
                info["exif"] = info["exif"][:p_value] + new_orientation + info["exif"][p_value + 2 :]
                break
    if info.get("xmp", None):
        xmp_data = info["xmp"].decode("utf-8")
        match = re.search(r'tiff:Orientation(="|>)([0-9])', xmp_data)
        if match:
            if original_orientation is None and int(match[2]) != 1:
                original_orientation = int(match[2])
            xmp_data = re.sub(r'tiff:Orientation="([0-9])"', "", xmp_data)
            xmp_data = re.sub(r"<tiff:Orientation>([0-9])</tiff:Orientation>", "", xmp_data)
            info["xmp"] = xmp_data.encode("utf-8")
    return original_orientation


def get_file_mimetype(fp) -> str:
    """Wrapper around `libheif.get_file_mimetype`

    :param fp: A filename (string), pathlib.Path object, file object or bytes.
       The file object must implement ``file.read``, ``file.seek`` and ``file.tell`` methods,
       and be opened in binary mode.
    :returns: "image/heic", "image/heif", "image/heic-sequence",
        "image/heif-sequence", "image/avif" or "image/avif-sequence"
    """

    __data = _get_bytes(fp, 50)
    return ffi.string(lib.heif_get_file_mime_type(__data, len(__data))).decode("utf-8")


def _get_bytes(fp, length=None) -> bytes:
    if isinstance(fp, (str, pathlib.Path)):
        with builtins.open(fp, "rb") as file:
            return file.read(length or -1)
    if hasattr(fp, "read"):
        offset = fp.tell() if hasattr(fp, "tell") else None
        result = fp.read(length or -1)
        if offset is not None and hasattr(fp, "seek"):
            fp.seek(offset)
        return result
    return bytes(fp)[:length]


def getxmp(xmp_data: bytes) -> dict:
    """Returns a dictionary containing the XMP tags.
    **Requires defusedxml to be installed.** Implementation taken from ``Pillow``.

    Used in :py:meth:`pillow_heif.as_opener._LibHeifImageFile.getxmp` for Pillow plugins.

    :param xmp_data: ``bytes`` containing string in UTF-8 encoding with XMP tags.

    :returns: XMP tags in a dictionary."""

    def get_name(tag):
        return tag.split("}")[1]

    def get_value(element):
        value = {get_name(k): v for k, v in element.attrib.items()}
        children = list(element)
        if children:
            for child in children:
                name = get_name(child.tag)
                child_value = get_value(child)
                if name in value:
                    if not isinstance(value[name], list):
                        value[name] = [value[name]]
                    value[name].append(child_value)
                else:
                    value[name] = child_value
        elif value:
            if element.text:
                value["text"] = element.text
        else:
            return element.text
        return value

    if xmp_data:
        if ElementTree is None:
            warn("XMP data cannot be read without defusedxml dependency")
            return {}
        _clear_data = xmp_data.rsplit(b"\x00", 1)
        if _clear_data[0]:
            root = ElementTree.fromstring(_clear_data[0])
            return {get_name(root.tag): get_value(root)}
    return {}



from os import getenv, path
from pathlib import Path
from shutil import copy
from subprocess import DEVNULL, PIPE, run
from sys import platform
from warnings import warn

from cffi import FFI

from libheif import build_libs

ffi = FFI()
with open("libheif/heif.h", "r", encoding="utf-8") as f:
    ffi.cdef(f.read())

ffi.cdef(
    """
    extern "Python" int64_t callback_tell(void*);
    extern "Python" int callback_seek(int64_t, void*);
    extern "Python" int callback_read(void*, size_t, void*);
    extern "Python" struct heif_error callback_write(struct heif_context*, const void*, size_t, void*);
    extern "Python" enum heif_reader_grow_status callback_wait_for_file_size(int64_t, void*);
"""
)

with open("pillow_heif/helpers.h", "r", encoding="utf-8") as f:
    ffi.cdef(f.read())


include_dirs = ["/usr/local/include", "/usr/include"]
library_dirs = ["/usr/local/lib", "/usr/lib", "/lib"]

include_path_prefix = ""
insert = False
if platform.lower() == "darwin":
    include_path_prefix = getenv("HOMEBREW_PREFIX")
    if not include_path_prefix:
        _result = run(["brew", "--prefix"], stderr=DEVNULL, stdout=PIPE, check=False)
        if not _result.returncode and _result.stdout is not None:
            include_path_prefix = _result.stdout.decode("utf-8").rstrip("\n")
    if not include_path_prefix:
        include_path_prefix = "/opt/local"
elif platform.lower() == "win32":
    include_path_prefix = getenv("VCPKG_PREFIX")
    if include_path_prefix is None:
        include_path_prefix = "C:\\msys64\\mingw64"
        warn(f"VCPKG_PREFIX environment variable is not set. Assuming `VCPKG_PREFIX={include_path_prefix}`")
else:
    include_path_prefix = build_libs.build_libs_linux()

# Need to include "lib" directory to find "heif" library.
include_path_prefix_lib = path.join(include_path_prefix, "lib")
if include_path_prefix_lib not in library_dirs:
    library_dirs.append(include_path_prefix_lib)

# MSYS2: rename "libheif.dll.a" to "libheif.lib"
if platform.lower() == "win32":
    lib_export_file = Path(path.join(include_path_prefix_lib, "libheif.dll.a"))
    if lib_export_file.is_file():
        copy(lib_export_file, path.join(include_path_prefix_lib, "libheif.lib"))
    else:
        warn("If you build this with MSYS2, you should not see this warning.")

# Adds project root to `include` path
include_dirs.append(path.dirname(path.dirname(path.abspath(__file__))))

ffi.set_source(
    "_pillow_heif_cffi",
    r"""
    #include "libheif/heif.h"
    #include "pillow_heif/helpers.c"
    """,
    include_dirs=include_dirs,
    library_dirs=library_dirs,
    libraries=["libheif"] if platform.lower() == "win32" else ["heif"],
    extra_compile_args=["/d2FH4-"] if platform.lower() == "win32" else [],
)

if __name__ == "__main__":
    ffi.compile(verbose=True)



import builtins
import os
from gc import collect
from io import BytesIO
from pathlib import Path
from unittest import mock

import dataset
import helpers
import pytest
from PIL import Image, ImageCms, ImageSequence, UnidentifiedImageError

import pillow_heif

os.chdir(os.path.dirname(os.path.abspath(__file__)))

pillow_heif.register_heif_opener()


@pytest.mark.parametrize("img_path", dataset.CORRUPTED_DATASET)
def test_heif_corrupted_open(img_path):
    for input_type in [img_path.read_bytes(), BytesIO(img_path.read_bytes()), img_path, builtins.open(img_path, "rb")]:
        try:
            pillow_heif.open_heif(input_type).load()
            assert False
        except pillow_heif.HeifError as exception:
            assert exception.code == pillow_heif.HeifErrorCode.INVALID_INPUT
            assert repr(exception).find("HeifErrorCode.INVALID_INPUT") != -1
            assert str(exception).find("Invalid input") != -1


@pytest.mark.parametrize("img_path", dataset.CORRUPTED_DATASET)
def test_pillow_corrupted_open(img_path):
    for input_type in [BytesIO(img_path.read_bytes()), img_path, builtins.open(img_path, "rb")]:
        with pytest.raises(UnidentifiedImageError):
            Image.open(input_type)


def test_heif_image_order():
    im = pillow_heif.open_heif(Path("images/heif/zPug_3.heic"))
    assert im.info["primary"] and im.primary_index() == 1
    assert not im[0].info["primary"]
    assert im[1].info["primary"]
    assert not im[2].info["primary"]


def test_pillow_image_order():
    im = Image.open(Path("images/heif/zPug_3.heic"))
    assert im.info["primary"] and im.tell() == 1
    im.seek(0)
    assert not im.info["primary"]
    im.seek(1)
    assert im.info["primary"]
    im.seek(2)
    assert not im.info["primary"]


@pytest.mark.parametrize("img_path", [Path("images/heif/RGBA_10.heif"), Path("images/heif/zPug_3.heic")])
def test_heif_inputs(img_path):
    def perform_test_input():
        with builtins.open(img_path, "rb") as fh:
            b = fh.read()
            bytes_io = BytesIO(b)
            for fp in [bytes_io, fh, img_path, img_path.as_posix(), b]:
                heif_file = pillow_heif.open_heif(fp)
                assert heif_file.mimetype
                assert min(heif_file.size) > 0
                assert heif_file.info
                assert getattr(heif_file[0], "_heif_ctx") is not None
                collect()
                # This will load all data
                for image in heif_file:
                    assert not getattr(image, "_img_data")
                    assert len(image.data) > 0
                    for thumbnail in image.thumbnails:
                        assert not getattr(thumbnail, "_img_data")
                        assert len(thumbnail.data) > 0
                        thumbnail.unload()
                    image.unload()
                collect()
                for image in heif_file:
                    assert not getattr(image, "_img_data")
                    assert len(image.data) > 0
                    for thumbnail in image.thumbnails:
                        assert not getattr(thumbnail, "_img_data")
                        assert len(thumbnail.data) > 0
                collect()
                assert getattr(heif_file[0], "_heif_ctx") is not None
                if not pillow_heif.options().ctx_in_memory:
                    assert getattr(heif_file[0]._heif_ctx, "fp") is not None
                    assert getattr(heif_file[0]._heif_ctx, "_fp_close_after") == isinstance(fp, (Path, str, bytes))
                # Create new heif_file
                heif_file_from = pillow_heif.HeifFile().add_from_heif(heif_file)
                collect()
                helpers.compare_heif_files_fields(heif_file_from, heif_file, ignore=["original_bit_depth"])
                for _ in heif_file:
                    _.unload()
                for _ in heif_file_from:
                    _.unload()
                collect()
                helpers.compare_heif_files_fields(heif_file_from, heif_file, ignore=["original_bit_depth"])
                heif_file = None  # noqa
                assert len(heif_file_from[len(heif_file_from) - 1].data)
                if not isinstance(fp, (Path, str, bytes)):
                    assert not fp.closed

    perform_test_input()
    try:
        pillow_heif.options().ctx_in_memory = False
        perform_test_input()
    finally:
        pillow_heif.options().ctx_in_memory = True


@pytest.mark.parametrize("img_path", [Path("images/heif/RGBA_10.heif"), Path("images/heif/zPug_3.heic")])
def test_pillow_inputs(img_path):
    with builtins.open(img_path, "rb") as fh:
        bytes_io = BytesIO(fh.read())
        for fp in [bytes_io, fh, img_path, img_path.as_posix()]:
            pillow_image = Image.open(fp)
            assert getattr(pillow_image, "fp") is not None
            pillow_image.load()
            for frame in ImageSequence.Iterator(pillow_image):
                assert len(frame.tobytes()) > 0
            heif_image = pillow_heif.from_pillow(pillow_image)
            helpers.compare_heif_to_pillow_fields(heif_image, pillow_image)
            assert len(pillow_heif.from_pillow(pillow_image, load_one=True)) == 1
            if getattr(pillow_image, "n_frames") > 1:
                assert getattr(pillow_image, "fp") is not None
            else:
                assert getattr(pillow_image, "fp") is None
            if not isinstance(fp, (Path, str)):
                assert not fp.closed


def test_pillow_after_load():
    img = Image.open(Path("images/heif/RGBA_10.heif"))
    assert getattr(img, "heif_file") is not None
    for i in range(3):
        img.load()
        collect()
        assert not getattr(img, "is_animated")
        assert getattr(img, "n_frames") == 1
        assert not img.info["thumbnails"]
        assert getattr(img, "heif_file") is None
        assert len(ImageSequence.Iterator(img)[0].tobytes())
    img = Image.open(Path("images/heif/zPug_3.heic"))
    for i in range(3):
        collect()
        assert getattr(img, "is_animated")
        assert getattr(img, "n_frames") == 3
        assert len(img.info["thumbnails"]) == 0 if i else 1
        assert getattr(img, "heif_file") is not None
        assert len(ImageSequence.Iterator(img)[0].info["thumbnails"]) == 2
        assert len(ImageSequence.Iterator(img)[1].info["thumbnails"]) == 1
        assert len(ImageSequence.Iterator(img)[2].info["thumbnails"]) == 0
        assert len(ImageSequence.Iterator(img)[0].tobytes())
        assert len(ImageSequence.Iterator(img)[1].tobytes())
        assert len(ImageSequence.Iterator(img)[2].tobytes())


@pytest.mark.parametrize("img_path", dataset.MINIMAL_DATASET)
def test_heif_from_heif(img_path):
    def heif_from_heif(hdr_to_8bit=True):
        heif_file = pillow_heif.open_heif(img_path, convert_hdr_to_8bit=hdr_to_8bit)
        collect()
        # Create Heif from Heif
        heif_file_from = pillow_heif.HeifFile().add_from_heif(heif_file)
        collect()
        # Create Heif from created Heif
        heif_file_from_from = pillow_heif.HeifFile().add_from_heif(heif_file_from)
        for _ in heif_file:
            _.unload()
        collect()
        helpers.compare_heif_files_fields(heif_file, heif_file_from, ignore=["original_bit_depth"])
        # Closing original Heif must not affect data in others two
        heif_file = None  # noqa
        for _ in heif_file_from:
            _.unload()
        for _ in heif_file_from_from:
            _.unload()
        collect()
        heif_file_from.load(everything=True)
        heif_file_from_from.load(everything=True)
        helpers.compare_heif_files_fields(heif_file_from, heif_file_from_from, ignore=["original_bit_depth"])
        heif_file_from = None  # noqa
        assert len(heif_file_from_from[len(heif_file_from_from) - 1].data)

    heif_from_heif(hdr_to_8bit=True)
    heif_from_heif(hdr_to_8bit=False)


@pytest.mark.parametrize("image_path", dataset.MINIMAL_DATASET)
def test_to_from_pillow(image_path):
    heif_file = pillow_heif.open_heif(image_path)
    images_list = [i.to_pillow() for i in heif_file]
    for i, image in enumerate(heif_file):
        helpers.compare_heif_to_pillow_fields(image, images_list[i])
    heif_from_pillow = pillow_heif.HeifFile()
    for image in images_list:
        heif_from_pillow.add_from_pillow(image)
    helpers.compare_heif_files_fields(heif_file, heif_from_pillow, ignore=["original_bit_depth"])


@pytest.mark.parametrize("image_path", dataset.FULL_DATASET)
def test_heif_read_images(image_path):
    def test_read_image(convert_hdr_to_8bit: bool) -> bool:
        heif_file = pillow_heif.open_heif(image_path, convert_hdr_to_8bit=convert_hdr_to_8bit)
        for image in heif_file:
            assert isinstance(pillow_heif.getxmp(image.info["xmp"]), dict)
            assert min(image.size) > 0
            assumed_mode = "RGBA" if image.has_alpha else "RGB"
            if image.bit_depth > 8:
                assumed_mode += f";{image.bit_depth}"
            assert image.mode == assumed_mode
            assert image.bit_depth >= 8
            minimal_stride = image.size[0] * 4 if image.has_alpha else image.size[0] * 3
            if image.bit_depth > 8:
                minimal_stride *= 2
            assert image.stride >= minimal_stride
            assert len(image.data) == image.stride * image.size[1]
            # This will load thumbnails too
            assert isinstance(image.load(), pillow_heif.HeifImage)
            for thumbnail in image.thumbnails:
                minimal_stride = thumbnail.size[0] * 4 if thumbnail.has_alpha else thumbnail.size[0] * 3
                if thumbnail.bit_depth > 8:
                    minimal_stride *= 2
                assert thumbnail.stride >= minimal_stride
                assert len(thumbnail.data) == thumbnail.stride * thumbnail.size[1]
                assert isinstance(thumbnail.load(), pillow_heif.HeifThumbnail)
        return heif_file.bit_depth > 8

    one_more = test_read_image(False)
    if one_more:
        test_read_image(True)
    try:
        pillow_heif.options().ctx_in_memory = False
        one_more = test_read_image(False)
        if one_more:
            test_read_image(True)
    finally:
        pillow_heif.options().ctx_in_memory = True


@pytest.mark.parametrize("image_path", dataset.FULL_DATASET)
def test_pillow_read_images(image_path):
    def test_read_image():
        pillow_image = Image.open(image_path)
        assert getattr(pillow_image, "fp") is not None
        assert getattr(pillow_image, "heif_file") is not None
        assert not getattr(pillow_image, "_close_exclusive_fp_after_loading")
        pillow_image.verify()
        images_count = len([_ for _ in ImageSequence.Iterator(pillow_image)])
        for i, image in enumerate(ImageSequence.Iterator(pillow_image)):
            assert image.info
            assert image.custom_mimetype in ("image/heic", "image/heif", "image/heif-sequence", "image/avif")
            if "icc_profile" in image.info and len(image.info["icc_profile"]) > 0:
                ImageCms.getOpenProfile(BytesIO(pillow_image.info["icc_profile"]))
            collect()
            assert len(ImageSequence.Iterator(pillow_image)[i].tobytes())
            for thumb in ImageSequence.Iterator(pillow_image)[i].info["thumbnails"]:
                if images_count > 1:
                    assert thumb.data is not None
                else:
                    assert thumb.data is None
            assert isinstance(image.getxmp(), dict)
        if images_count > 1:
            assert getattr(pillow_image, "fp") is not None
            assert getattr(pillow_image, "heif_file") is not None
            assert not getattr(pillow_image, "_close_exclusive_fp_after_loading")
        else:
            assert getattr(pillow_image, "fp") is None
            assert getattr(pillow_image, "heif_file") is None
            assert getattr(pillow_image, "_close_exclusive_fp_after_loading")
            # Testing here one more time, just for sure, that missing `heif_file` not affect anything.
            collect()
            assert pillow_image.tobytes()
            assert len(ImageSequence.Iterator(pillow_image)[0].tobytes())

    test_read_image()
    try:
        pillow_heif.options().ctx_in_memory = False
        test_read_image()
    finally:
        pillow_heif.options().ctx_in_memory = True


@pytest.mark.parametrize("img_path", dataset.TRUNCATED_DATASET)
def test_pillow_truncated_fail(img_path):
    truncated_heif = Image.open(img_path)
    with pytest.raises(pillow_heif.HeifError):
        truncated_heif.load()


@mock.patch("PIL.ImageFile.LOAD_TRUNCATED_IMAGES", True)
@pytest.mark.parametrize("img_path", dataset.TRUNCATED_DATASET)
def test_pillow_truncated_ok(img_path):
    im = Image.open(img_path)
    im.load()


def test_heif_index():
    heif_file = pillow_heif.open_heif(Path("images/heif/zPug_3.heic"))
    with pytest.raises(IndexError):
        heif_file[-1].load()
    with pytest.raises(IndexError):
        heif_file[len(heif_file)].load()
    with pytest.raises(IndexError):
        heif_file[0].thumbnails[len(heif_file[0].thumbnails)].load()
    with pytest.raises(IndexError):
        del heif_file[-1]
    with pytest.raises(IndexError):
        del heif_file[len(heif_file)]


@mock.patch("pillow_heif.misc.ElementTree", None)
def test_no_defusedxml(monkeypatch):
    with pytest.warns(UserWarning):
        pillow_heif.getxmp(b"xmp_data")


def test_read_heif():
    heif_file = pillow_heif.read_heif(Path("images/heif/zPug_3.heic"))
    for im in heif_file:
        assert im._img_data
        for thumbnail in im.thumbnails:
            assert not thumbnail._img_data
            thumbnail.load()
            assert thumbnail._img_data


def test_heif_etc():
    heif_file = pillow_heif.open_heif(Path("images/heif/zPug_3.heic"))
    heif_file.load(everything=False)
    assert getattr(heif_file[1], "_img_data")
    assert not getattr(heif_file[0], "_img_data")
    assert not getattr(heif_file[2], "_img_data")
    assert heif_file.size == heif_file[1].size
    assert heif_file.mode == heif_file[1].mode
    assert len(heif_file.data) == len(heif_file[1].data)
    assert heif_file.stride == heif_file[1].stride
    assert heif_file.has_alpha == heif_file[1].has_alpha
    assert heif_file.bit_depth == heif_file[1].bit_depth
    assert heif_file.original_bit_depth == heif_file[1].original_bit_depth


def test_heif_only_image_reference():
    empty_heif_container = pillow_heif.HeifFile()
    empty_heif_container.add_from_heif(pillow_heif.open_heif(Path("images/heif/zPug_3.heic"))[0])
    empty_heif_container.add_from_heif(pillow_heif.open_heif(Path("images/heif/zPug_3.heic"))[2])
    assert len(empty_heif_container) == 2
    empty_heif_container.load(everything=True)


@pytest.mark.parametrize(
    "im_path,original_path",
    (
        ("images/heif/L_10.heif", "images/non_heif/L_16.png"),
        ("images/heif/L_12.heif", "images/non_heif/L_16.png"),
        ("images/heif/RGB_10.heif", "images/non_heif/RGB_16.png"),
        ("images/heif/RGB_12.heif", "images/non_heif/RGB_16.png"),
        ("images/heif/RGBA_10.heif", "images/non_heif/RGBA_16.png"),
        ("images/heif/RGBA_12.heif", "images/non_heif/RGBA_16.png"),
    ),
)
def test_hdr_read(im_path, original_path):
    helpers.compare_hashes([im_path, original_path], max_difference=1)


@pytest.mark.parametrize(
    "im_path,original_path",
    (
        ("images/heif/L_10.avif", "images/non_heif/L_16.png"),
        ("images/heif/L_12.avif", "images/non_heif/L_16.png"),
        ("images/heif/RGB_10.avif", "images/non_heif/RGB_16.png"),
        ("images/heif/RGB_12.avif", "images/non_heif/RGB_16.png"),
        ("images/heif/RGBA_10.avif", "images/non_heif/RGBA_16.png"),
        ("images/heif/RGBA_12.avif", "images/non_heif/RGBA_16.png"),
    ),
)
@pytest.mark.skipif(not pillow_heif.options().avif, reason="requires AVIF decoder.")
def test_hdr_read_avif(im_path, original_path):
    helpers.compare_hashes([im_path, original_path], max_difference=1)



import os
from io import SEEK_END, BytesIO
from pathlib import Path

import pytest
from helpers import create_heif, hevc_enc
from PIL import Image, UnidentifiedImageError

from pillow_heif import from_pillow, open_heif, options, register_heif_opener

register_heif_opener()
os.chdir(os.path.dirname(os.path.abspath(__file__)))


@pytest.mark.skipif(not options().avif, reason="missing AV1 codecs.")
def test_avif_cfg_option():
    # when `options.avif` is False, pillow_heif must refuse to open `avif` files
    try:
        options().avif = False
        with pytest.raises(UnidentifiedImageError):
            Image.open(Path("images/heif/L_10.avif"))
    finally:
        options().reset()
    Image.open(Path("images/heif/L_10.avif"))


def test_strict_cfg_option():
    # libheif don't fully support burst images, so with strict=True pillow_heif should refuse them
    image_path = Path("images/heif_other/nokia/bird_burst.heic")
    try:
        options().update(strict=True)
        with pytest.raises(UnidentifiedImageError):
            Image.open(image_path)
        options().update(strict=False)
        assert getattr(Image.open(image_path), "n_frames") == 4
    finally:
        options().reset()


@pytest.mark.skipif(not hevc_enc(), reason="No HEVC encoder.")
def test_thumbnails_option():
    heif_buf = create_heif((128, 128), [64])
    try:
        heif_file = open_heif(heif_buf)
        assert heif_file.thumbnails
        assert str(heif_file.thumbnails[0]).find("bytes image data") == -1
        assert heif_file.thumbnails[0].data
        assert str(heif_file.thumbnails[0]).find("bytes image data") != -1
        # disabling thumbnails and checking them not to be present
        options().thumbnails = False
        heif_file = open_heif(heif_buf)
        assert not heif_file.thumbnails
    finally:
        options().reset()


@pytest.mark.skipif(not hevc_enc(), reason="No HEVC encoder.")
def test_quality_option():
    try:
        image = from_pillow(Image.linear_gradient(mode="L"))
        options().quality = 10
        out_buf_1_q10 = BytesIO()
        image.save(out_buf_1_q10)
        options().quality = 30
        # passing manual `quality` has higher priority then one in `options`
        out_buf_2_q10 = BytesIO()
        image.save(out_buf_2_q10, quality=10)
        assert out_buf_1_q10.seek(0, SEEK_END) == out_buf_2_q10.seek(0, SEEK_END)
        # image with quality 30% must have bigger size then 10%
        out_buf_3_q30 = BytesIO()
        image.save(out_buf_3_q30)
        assert out_buf_1_q10.seek(0, SEEK_END) < out_buf_3_q30.seek(0, SEEK_END)
    finally:
        options().reset()



import builtins
import os
import sys
from pathlib import Path
from platform import machine

import dataset
import pytest

import pillow_heif

os.chdir(os.path.dirname(os.path.abspath(__file__)))


def test_libheif_info():
    info = pillow_heif.libheif_info()
    assert info["version"]["libheif"] == "1.12.0"
    assert info["decoders"]["HEVC"]
    if machine().find("armv7") != -1:
        return
    assert info["decoders"]["AV1"]
    assert info["encoders"]["AV1"]
    if sys.maxsize > 2**32:
        assert info["encoders"]["HEVC"]


@pytest.mark.parametrize("img_path", dataset.FULL_DATASET)
def test_get_file_mimetype(img_path):
    mimetype = pillow_heif.get_file_mimetype(img_path)
    assert mimetype in ("image/heic", "image/heic-sequence", "image/heif", "image/heif-sequence", "image/avif")
    assert mimetype == pillow_heif.open_heif(img_path, False).mimetype


@pytest.mark.parametrize("img_path", dataset.FULL_DATASET)
def test_heif_check_filetype(img_path):
    with builtins.open(img_path, "rb") as fh:
        assert pillow_heif.check_heif(fh) != pillow_heif.HeifFiletype.NO
        assert pillow_heif.is_supported(fh)


def test_heif_str():
    str_img_nl_1 = "<HeifImage 64x64 RGB with no image data and 2 thumbnails>"
    str_img_nl_2 = "<HeifImage 64x64 RGB with no image data and 1 thumbnails>"
    str_img_nl_3 = "<HeifImage 96x64 RGB with no image data and 0 thumbnails>"
    str_img_l_1 = "<HeifImage 64x64 RGB with 12288 bytes image data and 2 thumbnails>"
    str_img_l_2 = "<HeifImage 64x64 RGB with 12288 bytes image data and 1 thumbnails>"
    str_thumb_nl = "<HeifThumbnail 32x32 RGB with no image data>"
    str_thumb_l = "<HeifThumbnail 32x32 RGB with 6144 bytes image data>"
    heif_file = pillow_heif.open_heif(Path("images/heif/zPug_3.heic"))
    assert str(heif_file) == f"<HeifFile with 3 images: ['{str_img_nl_1}', '{str_img_nl_2}', '{str_img_nl_3}']>"
    assert str(heif_file[0]) == str_img_nl_1
    assert str(heif_file[1]) == str_img_nl_2
    assert str(heif_file[2]) == str_img_nl_3
    assert str(heif_file.thumbnails[0]) == f"{str_thumb_nl} Original:{str_img_nl_2}"
    heif_file.load()
    assert str(heif_file) == f"<HeifFile with 3 images: ['{str_img_nl_1}', '{str_img_l_2}', '{str_img_nl_3}']>"
    heif_file.thumbnails[0].load()
    assert str(heif_file.thumbnails[0]) == f"{str_thumb_l} Original:{str_img_l_2}"
    heif_file = pillow_heif.HeifFile().add_from_heif(heif_file[0])
    assert str(heif_file) == f"<HeifFile with 1 images: ['{str_img_l_1}']>"
    assert str(heif_file.thumbnails[0]) == f"{str_thumb_nl} Original:{str_img_l_1}"
    heif_file.thumbnails[0].load()  # Should not change anything, thumbnails are cloned without data.
    assert str(heif_file.thumbnails[0]) == f"{str_thumb_nl} Original:{str_img_l_1}"



"""
Enums from libheif that can be used.
"""

from enum import IntEnum


class HeifChroma(IntEnum):
    """Chroma subsampling definitions."""

    UNDEFINED = 99
    """Undefined chroma."""
    MONOCHROME = 0
    """Mono chroma."""
    CHROMA_420 = 1
    """``Cb`` and ``Cr`` are each subsampled at a factor of 2 both horizontally and vertically."""
    CHROMA_422 = 2
    """The two chroma components are sampled at half the horizontal sample rate of luma."""
    CHROMA_444 = 3
    """Each of the three Y'CbCr components has the same sample rate."""
    INTERLEAVED_RGB = 10
    """Simple interleaved RGB."""
    INTERLEAVED_RGBA = 11
    """Interleaved RGB with Alpha channel."""
    INTERLEAVED_RRGGBB_BE = 12
    """10 bit RGB BE."""
    INTERLEAVED_RRGGBBAA_BE = 13
    """10 bit RGB BE with Alpha channel."""
    INTERLEAVED_RRGGBB_LE = 14
    """10 bit RGB LE."""
    INTERLEAVED_RRGGBBAA_LE = 15
    """10 bit RGB LE with Alpha channel."""


class HeifColorspace(IntEnum):
    """Colorspace format of the image."""

    UNDEFINED = 99
    """Undefined colorspace."""
    YCBCR = 0
    """https://en.wikipedia.org/wiki/YCbCr"""
    RGB = 1
    """RGB colorspace."""
    MONOCHROME = 2
    """Monochrome colorspace."""


class HeifChannel(IntEnum):
    """Type of color channel."""

    Y = 0
    """Luma component"""
    CB = 1
    """Blue difference"""
    CR = 2
    """Red difference"""
    R = 3
    """Red color channel"""
    G = 4
    """Green color channel"""
    B = 5
    """Blue color channel"""
    ALPHA = 6
    """Alpha color channel"""
    INTERLEAVED = 10
    """Interleaved color channel"""


def encode_fourcc(fourcc):
    """Encodes 4 bytes in reverse order"""
    return ord(fourcc[0]) << 24 | ord(fourcc[1]) << 16 | ord(fourcc[2]) << 8 | ord(fourcc[3])


class HeifColorProfileType(IntEnum):
    """
    Color profile type definitions.
    If there is an ICC profile and an NCLX profile, the ICC profile prioritized.
    """

    NOT_PRESENT = 0
    """There is no color profile."""
    NCLX = encode_fourcc("nclx")
    """ISO/IEC 29199-2:2020"""
    RICC = encode_fourcc("rICC")
    """Restricted ICC. ISO/IEC 14496-12:2022"""
    PROF = encode_fourcc("prof")
    """Usual ICC profile."""


class HeifFiletype(IntEnum):
    """Result of :py:func:`~pillow_heif.check_heif` function."""

    NO = 0
    """it is unsupported file"""
    YES_SUPPORTED = 1
    """it is heif and can be read by libheif"""
    YES_UNSUPPORTED = 2
    """it is heif, but probably cannot be read by libheif"""
    MAYBE = 3
    """not sure whether it is heif and is it supported"""


class HeifErrorCode(IntEnum):
    """Possible libheif error codes in :py:class:`~pillow_heif.HeifError`"""

    OK = 0
    """Everything ok, no error occurred."""
    INPUT_DOES_NOT_EXIST = 1
    """Input file does not exist."""
    INVALID_INPUT = 2
    """Error in input file. Corrupted or invalid content."""
    UNSUPPORTED_FILETYPE = 3
    """Input file type is not supported."""
    UNSUPPORTED_FEATURE = 4
    """Image requires an unsupported decoder feature."""
    USAGE_ERROR = 5
    """Library API has been used in an invalid way."""
    MEMORY_ALLOCATION_ERROR = 6
    """Could not allocate enough memory."""
    DECODER_PLUGIN_ERROR = 7
    """The decoder plugin generated an error."""
    ENCODER_PLUGIN_ERROR = 8
    """The encoder plugin generated an error."""
    ENCODING_ERROR = 9
    """Error during encoding or when writing to the output."""
    COLOR_PROFILE_DOES_NOT_EXIST = 10
    """Application has asked for a color profile type that does not exist."""


class HeifCompressionFormat(IntEnum):
    """Possible HEIF compression formats."""

    UNDEFINED = 0
    """The compression format is not defined."""
    HEVC = 1
    """The compression format is HEVC."""
    AVC = 2
    """The compression format is AVC."""
    JPEG = 3
    """The compression format is JPEG."""
    AV1 = 4
    """The compression format is AV1."""


# --------------------------------------------------------------------
# DEPRECATED CONSTANTS.
# pylint: disable=invalid-name

heif_filetype_no = 0
heif_filetype_yes_supported = 1
heif_filetype_yes_unsupported = 2
heif_filetype_maybe = 3



import os
from pathlib import Path

from PIL import Image, ImageSequence

import pillow_heif

pillow_heif.register_heif_opener()

os.chdir(os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))), "tests"))
TARGET_FOLDER = "../converted"

if __name__ == "__main__":
    os.makedirs(TARGET_FOLDER, exist_ok=True)
    image_path = Path("images/heif_other/nokia/alpha.heic")
    im = Image.open(image_path)
    result_path = os.path.join(TARGET_FOLDER, f"{image_path.stem}.heic")
    for frame in ImageSequence.Iterator(im):
        pillow_heif.add_thumbnails(frame, [256, 372])
    im.save(result_path, quality=35, save_all=True)

    # Next code will remove thumbnails
    im = Image.open(result_path)
    del im.info["thumbnails"][0]  # remove first thumbnail
    del im.info["thumbnails"][0]  # remove second thumbnail
    result_path = os.path.join(TARGET_FOLDER, f"{image_path.stem}_no_thumbs.heic")
    im.save(result_path, quality=35, save_all=True)



import gc
import sys
from io import BytesIO
from os import chdir, path
from pathlib import Path
from platform import machine

import helpers
import pytest
from PIL import Image

import pillow_heif

pytest.importorskip("pympler", reason="`pympler` not installed")
pytest.importorskip("numpy", reason="`numpy` not installed")

print(pillow_heif.libheif_info())

chdir(path.join(path.dirname(path.dirname(path.abspath(__file__))), "tests"))
pillow_heif.register_heif_opener()


def perform_open_save(iterations, image_path, do_save=True):
    for _ in range(iterations):
        image = Image.open(image_path)
        assert getattr(image, "heif_file") is not None
        if helpers.hevc_enc() and do_save:
            out_buf = BytesIO()
            image.save(out_buf, quality=20, format="HEIF", save_all=True)


@pytest.mark.skipif(sys.executable.lower().find("pypy") != -1, reason="Disabled on PyPy.")
@pytest.mark.parametrize("ctx_in_memory", (False, True))
def test_open_save_objects_leaks(ctx_in_memory):
    from pympler import summary, tracker

    pillow_heif.options().ctx_in_memory = ctx_in_memory
    image_path = Path("images/heif/zPug_3.heic")
    perform_open_save(1, image_path)
    gc.collect()
    _summary1 = tracker.SummaryTracker().create_summary()
    _summary1 = tracker.SummaryTracker().create_summary()  # noqa
    gc.collect()
    gc.set_debug(gc.DEBUG_SAVEALL)
    perform_open_save(5, image_path)
    gc.collect()
    summary2 = tracker.SummaryTracker().create_summary()
    results = summary._sweep(summary.get_diff(_summary1, summary2))  # noqa
    summary.print_(results)
    for result in results:
        # look for strings like: `_cffi_backend.__CDataGCP`, `_cffi_backend.__CDataOwnGC`
        assert result[0].find("cffi") == -1
        # look for strings like: `pillow_heif._libheif_ctx.LibHeifCtx`
        assert result[0].find("pillow_heif") == -1


def _get_mem_usage():
    from resource import RUSAGE_SELF, getpagesize, getrusage

    mem = getrusage(RUSAGE_SELF).ru_maxrss
    return mem * getpagesize() / 1024 / 1024


@pytest.mark.skipif(sys.platform.lower() == "win32", reason="requires Unix or macOS")
@pytest.mark.parametrize("ctx_in_memory", (False, True))
def test_open_save_leaks(ctx_in_memory):
    pillow_heif.options().ctx_in_memory = ctx_in_memory
    mem_limit = None
    for i in range(1000):
        # do_save=False
        # https://bitbucket.org/multicoreware/x265_git/issues/616/x265_encoder_open-leaks-memory-zoneparam
        perform_open_save(1, Path("images/heif/zPug_3.heic"), do_save=False)
        gc.collect()
        mem = _get_mem_usage()
        if i < 300:
            mem_limit = mem + 1
            continue
        assert mem <= mem_limit, f"memory usage limit exceeded after {i + 1} iterations"


@pytest.mark.skipif(sys.platform.lower() == "win32", reason="requires Unix or macOS")
@pytest.mark.skipif(machine().find("x86_64") == -1, reason="run only on x86_64")
def test_open_to_numpy_mem_leaks():
    import numpy as np

    mem_limit = None
    im_path = Path("images/heif/zPug_3.heic")
    for i in range(500):
        heif_file = pillow_heif.open_heif(im_path)
        _array = np.asarray(heif_file[0])  # noqa
        _array = None  # noqa
        gc.collect()
        mem = _get_mem_usage()
        if i < 300:
            mem_limit = mem + 1
            continue
        assert mem <= mem_limit, f"memory usage limit exceeded after {i + 1} iterations"



from gc import collect
from io import BytesIO

import pytest
from helpers import compare_hashes, hevc_enc
from PIL import Image, ImageSequence

import pillow_heif

pillow_heif.register_heif_opener()
if not hevc_enc():
    pytest.skip(reason="Requires HEIF encoder.", allow_module_level=True)


# Creating HEIF file in memory with 3 images.
# Second image is a Primary Image with EXIF and XMP data.
# First two images has 2 thumbnails each, third image has no thumbnails.
def create_thumbnail_heif(size):
    _ = Image.effect_mandelbrot(size, (-3, -2.5, 2, 2.5), 100)
    im_heif = pillow_heif.from_pillow(_)
    im_heif.add_from_pillow(_.crop((0, 0, 256, 256)))
    pillow_heif.add_thumbnails(im_heif[0], boxes=[128, 64])
    pillow_heif.add_thumbnails(im_heif[1], boxes=[128, 64])
    im_heif.add_from_pillow(_.crop((0, 0, 192, 192)))
    _heif_buf = BytesIO()
    exif = Image.Exif()
    exif[0x010E] = "this is a desc"
    im_heif.save(_heif_buf, primary_index=1, exif=exif.tobytes(), xmp=xmp_data)
    return _heif_buf


xmp_data = b"xmp_data"
heif_buf = create_thumbnail_heif((512, 512))


def test_heif_enumerate_thumbnails():
    heif_file = pillow_heif.open_heif(heif_buf)
    assert len(list(heif_file.thumbnails_all())) == 4
    assert len(list(heif_file.thumbnails_all(one_for_image=True))) == 2


def test_heif_enumerate_thumbnails_data():
    heif_file = pillow_heif.open_heif(heif_buf)
    for thumbnail in heif_file.thumbnails_all():
        assert len(thumbnail.data)
    for thumbnail in heif_file.thumbnails_all(one_for_image=True):
        assert len(thumbnail.data)


def test_heif_thumbnail():
    for i, img in enumerate(pillow_heif.open_heif(heif_buf)):
        thumbnail = pillow_heif.thumbnail(img)
        if i == 2:
            assert thumbnail.size == img.size
            assert isinstance(thumbnail, pillow_heif.HeifImage)
        else:
            assert thumbnail.size != img.size
            assert isinstance(thumbnail, pillow_heif.HeifThumbnail)
        assert len(thumbnail.data)


def test_pillow_thumbnail():
    for i, img in enumerate(ImageSequence.Iterator(Image.open(heif_buf))):
        thumbnail = pillow_heif.thumbnail(img)
        assert isinstance(thumbnail, Image.Image)
        if i == 2:
            assert thumbnail.size == img.size
            assert isinstance(thumbnail.info.get("thumbnails"), list)
        else:
            assert thumbnail.size != img.size
            assert thumbnail.info.get("thumbnails", None) is None
        assert thumbnail.info["primary"] == bool(i == 1)
        assert len(thumbnail.tobytes())


def test_pillow_thumbnail_image_loaded():
    for i, img in enumerate(ImageSequence.Iterator(Image.open(heif_buf))):
        img.load()
        thumbnail = pillow_heif.thumbnail(img)
        assert isinstance(thumbnail, Image.Image)
        if i == 2:
            assert thumbnail.size == img.size
            assert isinstance(thumbnail.info.get("thumbnails"), list)
        else:
            assert thumbnail.size != img.size
            assert thumbnail.info.get("thumbnails", None) is None
        assert thumbnail.info["primary"] == bool(i == 1)
        assert len(thumbnail.tobytes())


def test_pillow_thumbnail_one_image_loaded():
    heif_buf_one_image = BytesIO()
    Image.open(heif_buf).save(heif_buf_one_image, format="HEIF")
    im = Image.open(heif_buf_one_image)
    thumbnail = pillow_heif.thumbnail(im)
    assert thumbnail.size != im.size
    assert thumbnail.info.get("thumbnails") is None
    im.load()
    thumbnail = pillow_heif.thumbnail(im)
    assert thumbnail.size == im.size
    assert isinstance(thumbnail.info.get("thumbnails"), list)


def test_heif_thumbnail_above_size():
    for img in pillow_heif.open_heif(heif_buf):
        thumbnail = pillow_heif.thumbnail(img, min_box=9999)
        assert thumbnail.size == img.size
        assert isinstance(thumbnail, pillow_heif.HeifImage)
        assert len(thumbnail.data)


def test_pillow_thumbnail_above_size():
    for img in ImageSequence.Iterator(Image.open(heif_buf)):
        thumbnail = pillow_heif.thumbnail(img, min_box=9999)
        assert thumbnail.size == img.size
        assert isinstance(thumbnail, Image.Image)
        assert isinstance(thumbnail.info.get("thumbnails"), list)


def test_heif_thumbnail_below_size():
    for i, img in enumerate(pillow_heif.open_heif(heif_buf)):
        thumbnail = pillow_heif.thumbnail(img, min_box=16)
        if i == 2:
            assert thumbnail.size == img.size
            assert isinstance(thumbnail, pillow_heif.HeifImage)
        else:
            assert thumbnail.size != img.size
            assert isinstance(thumbnail, pillow_heif.HeifThumbnail)
        assert len(thumbnail.data)


def test_pillow_thumbnail_below_size():
    for i, img in enumerate(ImageSequence.Iterator(Image.open(heif_buf))):
        thumbnail = pillow_heif.thumbnail(img, min_box=16)
        if i == 2:
            assert thumbnail.size == img.size
            assert isinstance(thumbnail.info.get("thumbnails"), list)
        else:
            assert thumbnail.size != img.size
            assert thumbnail.info.get("thumbnails", None) is None
        assert len(thumbnail.tobytes())


def test_heif_thumbnail_primary():
    # checking if `thumbnail` return Primary Image thumbnail when input is a heif_file
    heif_file = pillow_heif.open_heif(heif_buf)
    thumbnail = pillow_heif.thumbnail(heif_file)
    assert thumbnail == heif_file[1].thumbnails[0]


def test_heif_thumbnail_no_xmp_exif():
    thumbnail = pillow_heif.thumbnail(pillow_heif.open_heif(heif_buf)[0])
    assert not thumbnail.info["exif"]
    assert not thumbnail.info["xmp"]


def test_pillow_thumbnail_no_xmp_exif():
    thumbnail = pillow_heif.thumbnail(ImageSequence.Iterator(Image.open(heif_buf))[0])
    assert not thumbnail.info["exif"]
    assert not thumbnail.info["xmp"]


def test_heif_thumbnail_xmp_exif():
    thumbnail = pillow_heif.thumbnail(pillow_heif.open_heif(heif_buf))
    assert thumbnail.info["exif"]
    assert thumbnail.info["xmp"] == xmp_data


def test_pillow_thumbnail_xmp_exif():
    thumbnail = pillow_heif.thumbnail(Image.open(heif_buf))
    assert thumbnail.info["exif"]
    assert thumbnail.info["xmp"] == xmp_data
    assert isinstance(thumbnail.getexif(), Image.Exif)


def test_heif_thumbnail_references():
    heif_file = pillow_heif.open_heif(heif_buf)
    thumbnails_all = list(heif_file.thumbnails_all())
    assert thumbnails_all[0].get_original() == heif_file[0]
    assert thumbnails_all[1].get_original() == heif_file[0]
    assert thumbnails_all[2].get_original() == heif_file[1]
    assert thumbnails_all[3].get_original() == heif_file[1]
    del heif_file[0]
    collect()
    assert thumbnails_all[0].get_original() is None
    assert thumbnails_all[0].get_original() is None
    assert str(thumbnails_all[0]).find("Original:None") != -1
    assert thumbnails_all[2].get_original() == heif_file[0]
    assert thumbnails_all[3].get_original() == heif_file[0]
    pillow_img = heif_file[0].to_pillow()
    assert pillow_img.info["thumbnails"][0].get_original() is None
    assert pillow_img.info["thumbnails"][1].get_original() is None
    img_from_pillow = pillow_heif.from_pillow(pillow_img)
    assert img_from_pillow.thumbnails[0].get_original() == img_from_pillow[0]
    assert img_from_pillow.thumbnails[1].get_original() == img_from_pillow[0]
    assert str(img_from_pillow.thumbnails[0]).find("Original:None") == -1


@pytest.mark.parametrize(
    "thumbs,expected_after",
    (
        (-1, [2, 2, 0]),
        ([-1], [2, 2, 0]),
        (0, [2, 2, 0]),
        ([0], [2, 2, 0]),
        (1, [2, 2, 0]),
        ([1], [2, 2, 0]),
        (64, [2, 2, 1]),
        ([64], [2, 2, 1]),
        ([2048], [2, 2, 0]),
        (96, [3, 3, 1]),
        ([96], [3, 3, 1]),
        ([84, 0], [3, 3, 1]),
        ([0, 84], [3, 3, 1]),
        ([96, 84], [4, 4, 2]),
    ),
)
@pytest.mark.parametrize("heif_file_buf", (heif_buf, create_thumbnail_heif((317, 311))))
@pytest.mark.parametrize("method", ("HeifFile", "HeifImage"))
def test_heif_add_thumbs(thumbs, expected_after, heif_file_buf, method):
    output = BytesIO()
    heif_file = pillow_heif.open_heif(heif_file_buf)
    if method == "HeifFile":
        pillow_heif.add_thumbnails(heif_file, thumbs)
    else:
        pillow_heif.add_thumbnails(heif_file[0], thumbs)
        pillow_heif.add_thumbnails(heif_file[1], thumbs)
        pillow_heif.add_thumbnails(heif_file[2], thumbs)
    heif_file.save(output, quality=-1)
    out_heif = pillow_heif.open_heif(output)
    for i in range(3):
        assert len(out_heif[i].thumbnails) == expected_after[i]
    compare_hashes([out_heif[0].to_pillow(), out_heif[0].thumbnails[0].to_pillow()], hash_size=8, max_difference=4)


@pytest.mark.parametrize(
    "thumbs,expected_after",
    (
        (-1, [2, 2, 0]),
        ([0], [2, 2, 0]),
        (64, [2, 2, 1]),
        ([2048], [2, 2, 0]),
        (96, [3, 3, 1]),
        ([0, 84], [3, 3, 1]),
        ([96, 84], [4, 4, 2]),
    ),
)
def test_pillow_add_thumbs(thumbs, expected_after):
    output = BytesIO()
    im = Image.open(heif_buf)
    for frame in ImageSequence.Iterator(im):
        pillow_heif.add_thumbnails(frame, thumbs)
    im.save(output, format="HEIF", quality=-1, save_all=True)
    out_heif = pillow_heif.open_heif(output)
    for i in range(3):
        assert len(out_heif[i].thumbnails) == expected_after[i]
    compare_hashes([out_heif[0].to_pillow(), out_heif[0].thumbnails[0].to_pillow()], hash_size=8, max_difference=4)


@pytest.mark.parametrize(
    "thumbs,expected_after",
    (
        (-1, 0),
        ([0], 0),
        (64, 1),
        ([2048], 0),
        ([0, 128], 1),
        ([64, 128], 2),
    ),
)
def test_pillow_add_thumbs_empty_info(thumbs, expected_after):
    output = BytesIO()
    im = Image.open(heif_buf)
    im.info.pop("thumbnails")
    pillow_heif.add_thumbnails(im, thumbs)
    im.save(output, format="HEIF", quality=-1)
    out_heif = pillow_heif.open_heif(output)
    assert len(out_heif.thumbnails) == expected_after
    if expected_after >= 1:
        compare_hashes([out_heif[0].to_pillow(), out_heif.thumbnails[0].to_pillow()], hash_size=8)


def test_heif_remove_thumbs():
    out_buffer = BytesIO()
    # removing first thumbnail of the first image.
    heif_file = pillow_heif.open_heif(heif_buf)
    del heif_file[0].thumbnails[0]
    heif_file.save(out_buffer)
    out_heif = pillow_heif.open_heif(out_buffer)
    assert len(out_heif[0].thumbnails) == 1
    assert len(out_heif[1].thumbnails) == 2
    assert len(out_heif[2].thumbnails) == 0
    # removing all thumbnails of the primary image.
    heif_file = pillow_heif.open_heif(heif_buf)
    heif_file.thumbnails.clear()
    heif_file.save(out_buffer)
    out_heif = pillow_heif.open_heif(out_buffer)
    assert len(out_heif[0].thumbnails) == 2
    assert len(out_heif[1].thumbnails) == 0
    assert len(out_heif[2].thumbnails) == 0


def test_pillow_remove_thumbs():
    out_buffer = BytesIO()
    # removing first thumbnail of the first image.
    im = Image.open(heif_buf)
    im.seek(0)
    del im.info["thumbnails"][0]
    im.save(out_buffer, format="HEIF", save_all=True)
    im_out = Image.open(out_buffer)
    assert len(ImageSequence.Iterator(im_out)[0].info["thumbnails"]) == 1
    assert len(ImageSequence.Iterator(im_out)[1].info["thumbnails"]) == 2
    assert len(ImageSequence.Iterator(im_out)[2].info["thumbnails"]) == 0
    # removing all thumbnails of the primary image.
    im = Image.open(heif_buf)
    im.info["thumbnails"].clear()
    im.save(out_buffer, format="HEIF", save_all=True)
    im_out = Image.open(out_buffer)
    assert len(ImageSequence.Iterator(im_out)[0].info["thumbnails"]) == 2
    assert len(ImageSequence.Iterator(im_out)[1].info["thumbnails"]) == 0
    assert len(ImageSequence.Iterator(im_out)[2].info["thumbnails"]) == 0



import sys
from os import chdir, environ, getcwd, makedirs, mkdir, path, remove
from platform import machine
from re import IGNORECASE, MULTILINE, search
from subprocess import DEVNULL, PIPE, STDOUT, CalledProcessError, TimeoutExpired, run

BUILD_DIR_PREFIX = environ.get("BUILD_DIR_PREFIX", "/tmp/pillow_heif")
BUILD_DIR_TOOLS = path.join(BUILD_DIR_PREFIX, "build-tools")
BUILD_DIR_LIBS = path.join(BUILD_DIR_PREFIX, "build-stuff")
INSTALL_DIR_LIBS = environ.get("INSTALL_DIR_LIBS", "/usr")


def download_file(url: str, out_path: str) -> bool:
    n_download_clients = 2
    for _ in range(2):
        try:
            run(
                ["wget", "-q", "--no-check-certificate", url, "-O", out_path],
                timeout=90,
                stderr=DEVNULL,
                stdout=DEVNULL,
                check=True,
            )
            return True
        except (CalledProcessError, TimeoutExpired):
            break
        except FileNotFoundError:
            n_download_clients -= 1
            break
    for _ in range(2):
        try:
            run(["curl", "-L", url, "-o", out_path], timeout=90, stderr=DEVNULL, stdout=DEVNULL, check=True)
            return True
        except (CalledProcessError, TimeoutExpired):
            break
        except FileNotFoundError:
            n_download_clients -= 1
            break
    if not n_download_clients:
        raise EnvironmentError("Both curl and wget cannot be found.")
    return False


def download_extract_to(url: str, out_path: str, strip: bool = True):
    makedirs(out_path)
    _archive_path = path.join(out_path, "download.tar.gz")
    download_file(url, _archive_path)
    _tar_cmd = f"tar -xf {_archive_path} -C {out_path}"
    if strip:
        _tar_cmd += " --strip-components 1"
    run(_tar_cmd.split(), check=True)
    remove(_archive_path)


def tool_check_version(name: str, min_version: str) -> bool:
    try:
        _ = run([name, "--version"], stdout=PIPE, stderr=DEVNULL, check=True)
    except (CalledProcessError, FileNotFoundError):
        return False
    if name == "nasm":
        _regexp = r"version\s*(\d+(\.\d+){2})"
    elif name == "autoconf":
        _regexp = r"(\d+(\.\d+){1})$"
    else:
        _regexp = r"(\d+(\.\d+){2})$"
    m_groups = search(_regexp, _.stdout.decode("utf-8"), flags=MULTILINE + IGNORECASE)
    if m_groups is None:
        return False
    current_version = tuple(map(int, str(m_groups.groups()[0]).split(".")))
    min_version = tuple(map(int, min_version.split(".")))
    if current_version >= min_version:
        print(f"Tool {name} with version {str(m_groups.groups()[0])} satisfy requirements.", flush=True)
        return True
    return False


def is_musllinux() -> bool:
    _ = run("ldd --version".split(), stdout=PIPE, stderr=STDOUT, check=False)
    if _.stdout and _.stdout.decode("utf-8").find("musl") != -1:
        return True
    return False


def build_tool_linux(url: str, name: str, min_version: str, configure_args: list = None, chmod=None):
    if min_version:
        if tool_check_version(name, min_version):
            return
    if configure_args is None:
        configure_args = []
    _tool_path = path.join(BUILD_DIR_TOOLS, name)
    if path.isdir(_tool_path):
        print(f"Cache found for {name}", flush=True)
        chdir(_tool_path)
    else:
        download_extract_to(url, _tool_path)
        chdir(_tool_path)
        if name == "cmake":
            run("./bootstrap -- -DCMAKE_USE_OPENSSL=OFF".split(), check=True)
        else:
            run(["./configure"] + configure_args, check=True)
        run("make".split(), check=True)
    run("make install".split(), check=True)
    run(f"{name} --version".split(), check=True)
    if chmod:
        run(f"chmod -R {chmod} {_tool_path}".split(), check=True)


def build_tools_linux(musl: bool = False):
    build_tool_linux(
        "https://pkg-config.freedesktop.org/releases/pkg-config-0.29.2.tar.gz",
        "pkg-config",
        "0.29.2" if not musl else "",
        configure_args=["--with-internal-glib"],
    )
    build_tool_linux("https://ftp.gnu.org/gnu/autoconf/autoconf-2.71.tar.gz", "autoconf", "2.71")
    build_tool_linux("https://ftp.gnu.org/gnu/automake/automake-1.16.5.tar.gz", "automake", "1.16.5")
    build_tool_linux("https://github.com/Kitware/CMake/archive/refs/tags/v3.22.3.tar.gz", "cmake", "3.16.1")
    build_tool_linux(
        "https://www.nasm.us/pub/nasm/releasebuilds/2.15.05/nasm-2.15.05.tar.gz", "nasm", "2.15.05", chmod="774"
    )


def is_library_installed(name: str) -> bool:
    if name.find("main") != -1 and name.find("reference") != -1:
        raise Exception("`name` param can not contain `main` and `reference` substrings.")
    _r = run(f"gcc -l{name}".split(), stdout=PIPE, stderr=STDOUT, check=False)
    if _r.stdout:
        _ = _r.stdout.decode("utf-8")
        if _.find("main") != -1 and _.find("reference") != -1:
            return True
    return False


def run_print_if_error(args) -> None:
    _ = run(args, stdout=PIPE, stderr=STDOUT, check=False)
    if _.returncode != 0:
        print(_.stdout.decode("utf-8"), flush=True)
        raise ChildProcessError(f"Failed: {args}")


def build_lib_linux(url: str, name: str, musl: bool = False):
    _lib_path = path.join(BUILD_DIR_LIBS, name)
    if path.isdir(_lib_path):
        print(f"Cache found for {name}", flush=True)
        chdir(path.join(_lib_path, "build")) if name == "aom" else chdir(_lib_path)
    else:
        _hide_build_process = False
        if name == "aom":
            _build_path = path.join(_lib_path, "build")
            makedirs(_build_path)
            download_extract_to(url, path.join(_lib_path, "aom"), False)
            chdir(_build_path)
        else:
            download_extract_to(url, _lib_path)
            chdir(_lib_path)
        if name == "libde265":
            run(["./autogen.sh"], check=True)
        print(f"Preconfiguring {name}...", flush=True)
        if name == "aom":
            cmake_args = "-DENABLE_TESTS=0 -DENABLE_TOOLS=0 -DENABLE_EXAMPLES=0 -DENABLE_DOCS=0".split()
            cmake_args += "-DENABLE_TESTDATA=0 -DCONFIG_AV1_ENCODER=1".split()
            cmake_args += "-DCMAKE_INSTALL_LIBDIR=lib -DBUILD_SHARED_LIBS=1".split()
            cmake_args += f"-DCMAKE_INSTALL_PREFIX={INSTALL_DIR_LIBS} ../aom".split()
            run(["cmake"] + cmake_args, check=True)
            _hide_build_process = True
        elif name == "x265":
            cmake_high_bits = "-DHIGH_BIT_DEPTH=ON -DEXPORT_C_API=OFF".split()
            cmake_high_bits += "-DENABLE_SHARED=OFF -DENABLE_CLI=OFF".split()
            mkdir("12bit")
            mkdir("10bit")
            chdir("10bit")
            run("cmake ./../source -DENABLE_HDR10_PLUS=ON".split() + cmake_high_bits, check=True)
            run_print_if_error("make -j4".split())
            run("mv libx265.a ../libx265_main10.a".split(), check=True)
            chdir("../12bit")
            run(["cmake"] + ["./../source", "-DMAIN12=ON"] + cmake_high_bits, check=True)
            run_print_if_error("make -j4".split())
            run("mv libx265.a ../libx265_main12.a".split(), check=True)
            chdir("..")
            cmake_args = f"-DCMAKE_INSTALL_PREFIX={INSTALL_DIR_LIBS} ./source".split()
            cmake_args += ["-G", "Unix Makefiles"]
            cmake_args += "-DLINKED_10BIT=ON -DLINKED_12BIT=ON -DEXTRA_LINK_FLAGS=-L.".split()
            cmake_args += "-DEXTRA_LIB='x265_main10.a;x265_main12.a'".split()
            run(["cmake"] + cmake_args, check=True)
            _hide_build_process = True
        else:
            configure_args = f"--prefix {INSTALL_DIR_LIBS}".split()
            if name == "libde265":
                configure_args += "--disable-sherlock265 --disable-dec265".split()
            elif name == "libheif":
                configure_args += "--disable-examples --disable-go".split()
            run(["./configure"] + configure_args, check=True)
        print(f"{name} configured. building...", flush=True)
        if _hide_build_process:
            run_print_if_error("make -j4".split())
        else:
            run("make -j4".split(), check=True)
        print(f"{name} build success.", flush=True)
    run("make install".split(), check=True)
    if musl:
        run(f"ldconfig {INSTALL_DIR_LIBS}/lib".split(), check=True)
    else:
        run("ldconfig", check=True)


def build_libs_linux() -> str:
    _install_flag = path.join(BUILD_DIR_PREFIX, "was_installed.flag")
    if path.isfile(_install_flag):
        print("Tools & Libraries already installed.", flush=True)
        return INSTALL_DIR_LIBS
    _is_musllinux = is_musllinux()
    _original_dir = getcwd()
    try:
        build_tools_linux(_is_musllinux)
        if sys.maxsize > 2**32:  # Build x265 encoder only on 64-bit systems.
            build_lib_linux(
                "https://bitbucket.org/multicoreware/x265_git/get/master.tar.gz",
                "x265",
                _is_musllinux,
            )
        build_lib_linux(
            "https://github.com/strukturag/libde265/releases/download/v1.0.8/libde265-1.0.8.tar.gz",
            "libde265",
            _is_musllinux,
        )
        if machine().find("armv7") == -1 and not is_library_installed("aom"):  # Are not trying to build aom on armv7.
            build_lib_linux("https://aomedia.googlesource.com/aom/+archive/v3.4.0.tar.gz", "aom", _is_musllinux)
        build_lib_linux(
            "https://github.com/strukturag/libheif/releases/download/v1.12.0/libheif-1.12.0.tar.gz",
            "libheif",
            _is_musllinux,
        )
        open(_install_flag, "w").close()
    finally:
        chdir(_original_dir)
    return INSTALL_DIR_LIBS



"""
Import all possible stuff that can be used.
"""


from ._lib_info import (
    have_decoder_for_format,
    have_encoder_for_format,
    libheif_info,
    libheif_version,
)
from ._options import options
from ._version import __version__
from .as_opener import HeifImageFile, register_heif_opener
from .constants import heif_filetype_maybe  # DEPRECATED
from .constants import heif_filetype_no  # DEPRECATED
from .constants import heif_filetype_yes_supported  # DEPRECATED
from .constants import heif_filetype_yes_unsupported  # DEPRECATED
from .constants import (
    HeifChannel,
    HeifChroma,
    HeifColorProfileType,
    HeifColorspace,
    HeifCompressionFormat,
    HeifErrorCode,
    HeifFiletype,
)
from .error import HeifError

# pylint: disable=redefined-builtin
from .heif import (
    HeifFile,
    HeifImage,
    HeifThumbnail,
    check,
    check_heif,
    from_bytes,
    from_pillow,
    is_supported,
    open,
    open_heif,
    read,
    read_heif,
)
from .misc import get_file_mimetype, getxmp, set_orientation
from .thumbnails import add_thumbnails, thumbnail
